---- MODULE MemoirLL1CorrectnessInvariants ----

(***************************************************************************)
(* This module defines the three correctness invariants needed to prove    *)
(* that the `^\basicname^' spec implements the high-level spec.            *)
(***************************************************************************)

EXTENDS MemoirLL1TypeSafety

----

(***************************************************************************)
(* The UnforgeabilityInvariant states that, for any authenticator residing *)
(* in the user's RAM, if the authenticator validates using the symmetric   *)
(* key in the NVRAM, the authenticator is in the set of authenticators     *)
(* that the user observed `^\name^' to produce.                            *)
(*                                                                         *)
(* This is a somewhat boring invariant.  It really just extends the        *)
(* assumption in the LL1CorruptRAM action, which constraints the set of    *)
(* authenticators the user can create.  If we had written the low-level    *)
(* spec differently, such that this constraint had been expressed in the   *)
(* LL1PerformOperation and LL1RepeatOperation actions instead of the       *)
(* LL1CorruptRAM action, this invariant might not be necessary.            *)
(***************************************************************************)

UnforgeabilityInvariant == \*`^\label{DEF-UnforgeabilityInvariant}^'
    \A historyStateBinding \in HashType :
        ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator) =>
            LL1RAM.authenticator \in LL1ObservedAuthenticators

----

(***************************************************************************)
(* The InclusionInvariant states that, for any history summary and input   *)
(* that together hash to the history summary in the NVRAM, if this history *)
(* summary is bound to some public and private state by an authenticated   *)
(* history state binding, then the result of invoking the service with     *)
(* this public state, private state, and input will yield an output that   *)
(* is already in the set of observed outputs and a new history state       *)
(* binding that is already authenticated.                                  *)
(*                                                                         *)
(* This invariant is needed to show that the LL1RepeatOperation action     *)
(* does not have any ill effects.  In particular, this invariant tells us  *)
(* that the output that LL1RepeatOperation will produce is already in      *)
(* LL1ObservedOutputs, and the new authenticator that LL1RepeatOperation   *)
(* will produce makes an assertion that is already being asserted by some  *)
(* authenticator in LL1ObservedAuthenticators.                             *)
(***************************************************************************)

InclusionInvariant == \*`^\label{DEF-InclusionInvariant}^'
    \A  input \in InputType,
        historySummary \in HashType,
        publicState \in PublicStateType,
        privateStateEnc \in PrivateStateEncType :
        LET
            stateHash == Hash(publicState, privateStateEnc)
            historyStateBinding == Hash(historySummary, stateHash)
            privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, privateStateEnc)
            sResult == Service(publicState, privateState, input)
            newPrivateStateEnc ==
                SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
            newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
            newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
        IN
              ( /\  LL1NVRAM.historySummary = Hash(historySummary, input)
                /\  LL1HistoryStateBindingAuthenticated(historyStateBinding) )
            =>
              ( /\  sResult.output \in LL1ObservedOutputs
                /\  LL1HistoryStateBindingAuthenticated(newHistoryStateBinding) )

----

(***************************************************************************)
(* The UniquenessInvariant states that the the history summary in the      *)
(* NVRAM is bound to only one public and private state by an authenticator *)
(* in the set of observed authenticators.  This invariant is used in       *)
(* several places in the implementation proof.                             *)
(*                                                                         *)
(* In the NonAdvancementLemma, the property of uniqueness is needed to     *)
(* show that the refined public and private state does not change when the *)
(* NVRAM and set of observed authenticators does not change.  If there     *)
(* were more than one state bound to the history summary in the NVRAM, a   *)
(* low-level stuttering step could lead to a high-level change in state.   *)
(*                                                                         *)
(* In the base case of the `^\basicname^' implementation proof, the        *)
(* uniqueness property is needed to show that the refined state hash       *)
(* corresponds uniquely to the initial state hash.                         *)
(*                                                                         *)
(* Within the induction of the `^\basicname^' implementation proof, in the *)
(* case of LL1PerformOperation, the uniqueness property is used in two     *)
(* places.  First, it is needed to show that the public and private state  *)
(* in the arguments to the service correspond to the refined high-level    *)
(* state.  Second, it is needed to show that the public and private state  *)
(* produced as the result from the service correspond to the refined       *)
(* high-level primed state.                                                *)
(***************************************************************************)

UniquenessInvariant == \*`^\label{DEF-UniquenessInvariant}^'
    \A stateHash1, stateHash2 \in HashType :
        LET
            historyStateBinding1 == Hash(LL1NVRAM.historySummary, stateHash1)
            historyStateBinding2 == Hash(LL1NVRAM.historySummary, stateHash2)
        IN
              ( /\  LL1HistoryStateBindingAuthenticated(historyStateBinding1)
                /\  LL1HistoryStateBindingAuthenticated(historyStateBinding2) )
            =>
                stateHash1 = stateHash2

----

(***************************************************************************)
(* Collectively, we refer to these three invariants as the correctness     *)
(* invariants for the `^\basicname^' implementation.                       *)
(***************************************************************************)

CorrectnessInvariants == \*`^\label{DEF-CorrectnessInvariants}^'
    /\ UnforgeabilityInvariant
    /\ InclusionInvariant
    /\ UniquenessInvariant

====
