---- MODULE MemoirLL1InvarianceLemmas ----

(***************************************************************************)
(* This module states and proves several lemmas that are useful for        *)
(* proving the `^\basicname^' invariance properties.                       *)
(*                                                                         *)
(* The lemmas in this module are:                                          *)
(*     SymmetricKeyConstantLemma                                           *)
(*     LL1NVRAMHistorySummaryUncorruptedUnchangedLemma                     *)
(*     LL1RepeatOperationUnchangedObseredOutputsLemma                      *)
(*     LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma   *)
(*     LL1RAMUnforgeabilityUnchangedLemma                                  *)
(*     LL1DiskUnforgeabilityUnchangedLemma                                 *)
(*     InclusionUnchangedLemma                                             *)
(*     CardinalityUnchangedLemma                                           *)
(*     UniquenessUnchangedLemma                                            *)
(*     UnchangedAuthenticatedHistoryStateBindingsLemma                     *)
(***************************************************************************)

EXTENDS MemoirLL1SupplementalInvariants

(***************************************************************************)
(* Proof relating to cardinality require some basic properties of          *)
(* inequalities on natural numbers.  The prover requires that we state     *)
(* these explicitly.                                                       *)
(***************************************************************************)

THEOREM LEQTransitive == \A n, m, q \in Nat : n <= m /\ m <= q => n <= q \*`^\label{THEOREM-LEQTransitive}^'
  OBVIOUS (*{ by (isabelle "(auto dest: nat_leq_trans)") }*)

THEOREM GEQorLT == \A n, m \in Nat : n >= m <=> ~(m > n) \*`^\label{THEOREM-GEQorLT}^'
  OBVIOUS (*{ by (isabelle "(auto simp: nat_not_less dest: nat_leq_less_trans)") }*)

----

(***************************************************************************)
(* The SymmetricKeyConstantLemma states that the LL1Next actions do not    *)
(* change the value of the symmetric key in NVRAM.  The proof follows      *)
(* directly from the definition of the actions.                            *)
(***************************************************************************)

THEOREM SymmetricKeyConstantLemma == \*`^\label{THEOREM-SymmetricKeyConstantLemma}^'
    [LL1Next]_LL1Vars => UNCHANGED LL1NVRAM.symmetricKey
  <1>1. HAVE [LL1Next]_LL1Vars
  <1>2. CASE UNCHANGED LL1Vars
    <2>1. UNCHANGED LL1NVRAM
      BY <1>2 DEF LL1Vars
    <2>2. QED
      BY <2>1
  <1>3. CASE LL1Next
    <2>1. CASE LL1MakeInputAvailable
      <3>1. UNCHANGED LL1NVRAM
        BY <2>1 DEF LL1MakeInputAvailable
      <3>2. QED
        BY <3>1, HashCardinalityAccumulative
    <2>2. CASE LL1PerformOperation
      <3>1. PICK input \in LL1AvailableInputs : LL1PerformOperation!(input)!1
        BY <2>2 DEF LL1PerformOperation
      <3>2. LL1NVRAM' = [ historySummary |-> LL1PerformOperation!(input)!newHistorySummary,
                       symmetricKey |-> LL1NVRAM.symmetricKey ]
        BY <3>1
      <3>3. LL1NVRAM.symmetricKey' = LL1NVRAM.symmetricKey
        BY <3>2
      <3>4. QED
        BY <3>3
    <2>3. CASE LL1RepeatOperation
      <3>1. UNCHANGED LL1NVRAM
        BY <2>3 DEF LL1RepeatOperation
      <3>2. QED
        BY <3>1
    <2>4. CASE LL1Restart
      <3>1. UNCHANGED LL1NVRAM
        BY <2>4 DEF LL1Restart
      <3>2. QED
        BY <3>1
    <2>5. CASE LL1ReadDisk
      <3>1. UNCHANGED LL1NVRAM
        BY <2>5 DEF LL1ReadDisk
      <3>2. QED
        BY <3>1
    <2>6. CASE LL1WriteDisk
      <3>1. UNCHANGED LL1NVRAM
        BY <2>6 DEF LL1WriteDisk
      <3>2. QED
        BY <3>1
    <2>7. CASE LL1CorruptRAM
      <3>1. UNCHANGED LL1NVRAM
        BY <2>7 DEF LL1CorruptRAM
      <3>2. QED
        BY <3>1
    <2>8. CASE LL1RestrictedCorruption
      <3>1. PICK garbageHistorySummary \in HashType :
                LL1RestrictedCorruption!nvram!(garbageHistorySummary)
        BY <2>8 DEF LL1RestrictedCorruption
      <3>2. LL1NVRAM' = [ historySummary |-> garbageHistorySummary,
                       symmetricKey |-> LL1NVRAM.symmetricKey ]
        BY <3>1
      <3>3. QED
        BY <3>2
    <2>9. QED
      BY <1>3, <2>1, <2>2, <2>3, <2>4, <2>5, <2>6, <2>7, <2>8 DEF LL1Next
  <1>4. QED
    BY <1>1, <1>2, <1>3

----

(***************************************************************************)
(* The LL1NVRAMHistorySummaryUncorruptedUnchangedLemma states that, if     *)
(* there are no changes to the NVRAM or to the authentication status of    *)
(* any history state binding, then the truth value of the                  *)
(* LL1NVRAMHistorySummaryUncorrupted predicate is unchanged.               *)
(***************************************************************************)

THEOREM LL1NVRAMHistorySummaryUncorruptedUnchangedLemma == \*`^\label{THEOREM-LL1NVRAMHistorySummaryUncorruptedUnchangedLemma}^'
      ( /\  LL1TypeInvariant
        /\  UNCHANGED LL1NVRAM
        /\  \A historyStateBinding1 \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1) )
    =>
        UNCHANGED LL1NVRAMHistorySummaryUncorrupted
  (*************************************************************************)
  (* We begin by assuming the antecedent.                                  *)
  (*************************************************************************)
  <1>1. HAVE /\  LL1TypeInvariant
             /\  UNCHANGED LL1NVRAM
             /\  \A historyStateBinding1 \in HashType :
                     UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
  (*************************************************************************)
  (* We first consider the case in which the truth value of the predicate  *)
  (* is true.                                                              *)
  (*************************************************************************)
  <1>2. CASE LL1NVRAMHistorySummaryUncorrupted = TRUE
    (***********************************************************************)
    (* To show that the value is unchanged, it suffices to show that the   *)
    (* value is true in the primed state.                                  *)
    (***********************************************************************)
    <2>1. SUFFICES
          ASSUME TRUE
          PROVE LL1NVRAMHistorySummaryUncorrupted' = TRUE
      BY <1>2
    (***********************************************************************)
    (* We pick some state hash for which the                               *)
    (* LL1NVRAMHistorySummaryUncorrupted is true in the unprimed state.    *)
    (***********************************************************************)
    <2>2. PICK stateHash \in HashType :
              LL1NVRAMHistorySummaryUncorrupted!(stateHash)!1
      BY <1>2 DEF LL1NVRAMHistorySummaryUncorrupted
    (***********************************************************************)
    (* We copy the definition from the LET in                              *)
    (* LL1NVRAMHistorySummaryUncorrupted.                                  *)
    (***********************************************************************)
    <2> historyStateBinding == Hash(LL1NVRAM.historySummary, stateHash)
    (***********************************************************************)
    (* The LL1NVRAMHistorySummaryUncorrupted predicate has two conditions. *)
    (* First, that there exists a state hash in HashType, for which we     *)
    (* have a witness.                                                     *)
    (***********************************************************************)
    <2>3. stateHash \in HashType
      BY <2>2
    (***********************************************************************)
    (* The second condition is that the history state binding is           *)
    (* authenticated in the primed state.                                  *)
    (***********************************************************************)
    <2>4. LL1HistoryStateBindingAuthenticated(historyStateBinding)'
      (*********************************************************************)
      (* We will use the assumption that there is no change to the         *)
      (* authentication status of any history state binding.               *)
      (*********************************************************************)
      <3>1. \A historyStateBinding1 \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
        BY <1>1
      (*********************************************************************)
      (* This requires that we prove the type of the history state         *)
      (* binding.                                                          *)
      (*********************************************************************)
      <3>2. historyStateBinding \in HashType
        <4>1. LL1NVRAM.historySummary \in HashDomain
          <5>1. LL1NVRAM.historySummary \in HashType
            <6>1. LL1TypeInvariant
              BY <1>1
            <6>2. QED
              BY <6>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>2. stateHash \in HashDomain
          BY <2>3 DEF HashDomain
        <4>3. QED
          BY <4>1, <4>2, HashTypeSafe DEF historyStateBinding
      (*********************************************************************)
      (* We know that the history state binding is authenticated in the    *)
      (* unprimed state.                                                   *)
      (*********************************************************************)
      <3>3. LL1HistoryStateBindingAuthenticated(historyStateBinding)
        BY <2>2
      (*********************************************************************)
      (* By applying the assumption that there is no change to the         *)
      (* authentication status of any history state binding, we show that  *)
      (* the history state binding is authenticated in the primed state.   *)
      (*********************************************************************)
      <3>4. QED
        BY <3>1, <3>2, <3>3
    (***********************************************************************)
    (* Since both conditions are satisfied, the                            *)
    (* LL1NVRAMHistorySummaryUncorrupted predicate is true in the primed   *)
    (* state.                                                              *)
    (***********************************************************************)
    <2>5. QED
      BY <2>3, <2>4 DEF LL1NVRAMHistorySummaryUncorrupted
  (*************************************************************************)
  (* We then consider the case in which the truth value of the predicate   *)
  (* is false.                                                             *)
  (*************************************************************************)
  <1>3. CASE LL1NVRAMHistorySummaryUncorrupted = FALSE
    (***********************************************************************)
    (* To show that the value is unchanged, it suffices to show that if    *)
    (* the value were unequal to false in the primed state, we would have  *)
    (* a contradiction.                                                    *)
    (***********************************************************************)
    <2>1. SUFFICES
          ASSUME LL1NVRAMHistorySummaryUncorrupted' # FALSE
          PROVE FALSE
      BY <1>3
    <2>2. LL1NVRAMHistorySummaryUncorrupted'
      <3>1. LL1NVRAMHistorySummaryUncorrupted' \in BOOLEAN
        BY DEF LL1NVRAMHistorySummaryUncorrupted
      <3>2. QED
        BY <2>1, <3>1
    (***********************************************************************)
    (* We pick some state hash for which the                               *)
    (* LL1NVRAMHistorySummaryUncorrupted is true in the primed state.      *)
    (***********************************************************************)
    <2>3. PICK stateHash \in HashType :
              LL1NVRAMHistorySummaryUncorrupted!(stateHash)!1'
      BY <2>2 DEF LL1NVRAMHistorySummaryUncorrupted
    (***********************************************************************)
    (* We copy the definition from the LET in                              *)
    (* LL1NVRAMHistorySummaryUncorrupted.                                  *)
    (***********************************************************************)
    <2> historyStateBinding == Hash(LL1NVRAM.historySummary, stateHash)
    (***********************************************************************)
    (* The LL1NVRAMHistorySummaryUncorrupted predicate has two conditions. *)
    (* First, that there exists a state hash in HashType, for which we     *)
    (* have a witness.                                                     *)
    (***********************************************************************)
    <2>4. stateHash \in HashType
      BY <2>3
    (***********************************************************************)
    (* The second condition is that the history state binding is           *)
    (* authenticated in the unprimed state.                                *)
    (***********************************************************************)
    <2>5. LL1HistoryStateBindingAuthenticated(historyStateBinding)
      (*********************************************************************)
      (* We will use the assumption that there is no change to the         *)
      (* authentication status of any history state binding.               *)
      (*********************************************************************)
      <3>1. \A historyStateBinding1 \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
        BY <1>1
      (*********************************************************************)
      (* This requires that we prove the type of the history state         *)
      (* binding.                                                          *)
      (*********************************************************************)
      <3>2. historyStateBinding \in HashType
        <4>1. LL1NVRAM.historySummary \in HashDomain
          <5>1. LL1NVRAM.historySummary \in HashType
            <6>1. LL1TypeInvariant
              BY <1>1
            <6>2. QED
              BY <6>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>2. stateHash \in HashDomain
          BY <2>4 DEF HashDomain
        <4>3. QED
          BY <4>1, <4>2, HashTypeSafe DEF historyStateBinding
      (*********************************************************************)
      (* We know, in our contradictory universe, that the history state    *)
      (* binding is authenticated in the primed state.                     *)
      (*********************************************************************)
      <3>3. LL1HistoryStateBindingAuthenticated(historyStateBinding)'
        BY <2>3
      (*********************************************************************)
      (* By applying the assumption that there is no change to the         *)
      (* authentication status of any history state binding, we show that  *)
      (* the history state binding is authenticated in the unprimed state. *)
      (*********************************************************************)
      <3>4. QED
        BY <3>1, <3>2, <3>3
    (***********************************************************************)
    (* Since both conditions are satisfied, the                            *)
    (* LL1NVRAMHistorySummaryUncorrupted predicate is true in the unprimed *)
    (* state.                                                              *)
    (***********************************************************************)
    <2>6. LL1NVRAMHistorySummaryUncorrupted
      BY <2>4, <2>5 DEF LL1NVRAMHistorySummaryUncorrupted
    (***********************************************************************)
    (* However, we are considering the case in which the                   *)
    (* LL1NVRAMHistorySummaryUncorrupted predicate is false in the         *)
    (* unprimed state, so we have a contradiction.                         *)
    (***********************************************************************)
    <2>7. QED
      BY <1>3, <2>6
  (*************************************************************************)
  (* The predicate has a boolean truth value.                              *)
  (*************************************************************************)
  <1>4. LL1NVRAMHistorySummaryUncorrupted \in BOOLEAN
    BY DEF LL1NVRAMHistorySummaryUncorrupted
  <1>5. QED
    BY <1>2, <1>3, <1>4

----

(***************************************************************************)
(* The LL1RepeatOperationUnchangedObseredOutputsLemma states that the      *)
(* LL1RepeatOperation action does not change the value of                  *)
(* LL1ObservedOutputs.  Together with the                                  *)
(* LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma       *)
(* below, this is the essence of why LL1RepeatOperation does not cause any *)
(* change to the refined high-level state.                                 *)
(***************************************************************************)

THEOREM LL1RepeatOperationUnchangedObseredOutputsLemma == \*`^\label{THEOREM-LL1RepeatOperationUnchangedObseredOutputsLemma}^'
    LL1TypeInvariant /\ UnforgeabilityInvariant /\ InclusionInvariant /\ LL1RepeatOperation =>
        UNCHANGED LL1ObservedOutputs
  (*************************************************************************)
  (* First, we assume the antecedents.                                     *)
  (*************************************************************************)
  <1>1. HAVE LL1TypeInvariant /\ UnforgeabilityInvariant /\ InclusionInvariant /\ LL1RepeatOperation
  (*************************************************************************)
  (* Then, we pick some input for which LL1RepeatOperation is true.        *)
  (*************************************************************************)
  <1>2. PICK input \in LL1AvailableInputs : LL1RepeatOperation!(input)!1 
    BY <1>1 DEF LL1RepeatOperation
  (*************************************************************************)
  (* To simplify the writing of the proof, we re-state some of the         *)
  (* definitions from the LL1RepeatOperation action.                       *)
  (*************************************************************************)
  <1> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
  <1> historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
  <1> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
  <1> sResult == Service(LL1RAM.publicState, privateState, input)
  (*************************************************************************)
  (* We then assert the type safety of these definitions, with the help of *)
  (* the LL1RepeatOperationDefsTypeSafeLemma.                              *)
  (*************************************************************************)
  <1>3. /\  stateHash \in HashType
        /\  historyStateBinding \in HashType
        /\  privateState \in PrivateStateType
        /\  sResult \in ServiceResultType
        /\  sResult.newPublicState \in PublicStateType
        /\  sResult.newPrivateState \in PrivateStateType
        /\  sResult.output \in OutputType
    <2>1. input \in LL1AvailableInputs
      BY <1>2
    <2>2. LL1TypeInvariant
      BY <1>1
    <2>3. QED
      BY <2>1, <2>2, LL1RepeatOperationDefsTypeSafeLemma
  (*************************************************************************)
  (* We hide the definitions, so they don't overwhelm the prover.  We'll   *)
  (* pull them in as necessary below.                                      *)
  (*************************************************************************)
  <1> HIDE DEF stateHash, historyStateBinding, privateState, sResult
  (*************************************************************************)
  (* From the definition of LL1RepeatOperation, we see the primed state of *)
  (* LL1ObservedOutputs is formed by unioning in the output from the       *)
  (* service.                                                              *)
  (*************************************************************************)
  <1>4. LL1ObservedOutputs' = LL1ObservedOutputs \union {sResult.output}
    BY <1>2 DEF LL1RepeatOperation, sResult, privateState
  (*************************************************************************)
  (* We then show that the output from the service is already in     *)
  (* LL1ObservedOutputs.                                                   *)
  (*************************************************************************)
  <1>5. sResult.output \in LL1ObservedOutputs
    (***********************************************************************)
    (* Our strategy is to use the InclusionInvariant.  We first have to    *)
    (* show that all of the types are satisfied.                           *)
    (***********************************************************************)
    <2>1. LL1TypeInvariant
      BY <1>1
    <2>2. input \in InputType
      <3>1. input \in LL1AvailableInputs
        BY <1>2
      <3>2. LL1AvailableInputs \subseteq InputType
        BY <2>1 DEF LL1TypeInvariant
      <3>3. QED
        BY <3>1, <3>2
    <2>3. /\  LL1RAM.historySummary \in HashType
          /\  LL1RAM.publicState \in PublicStateType
          /\  LL1RAM.privateStateEnc \in PrivateStateEncType
      BY <2>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    (***********************************************************************)
    (* We then have to show that the antecedents in the InclusionInvariant *)
    (* are satisfied.                                                      *)
    (***********************************************************************)
    <2>4. LL1NVRAM.historySummary = Hash(LL1RAM.historySummary, input)
      BY <1>2
    <2>5. LL1HistoryStateBindingAuthenticated(historyStateBinding)
      (*********************************************************************)
      (* To show that the history state binding is authenticated, we       *)
      (* demonstrate that LL1RAM.authenticator is a sufficient witness for *)
      (* the existential quantifier within the definition of               *)
      (* LL1HistoryStateBindingAuthenticated.                              *)
      (*********************************************************************)
      <3>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
        BY <1>2 DEF historyStateBinding, stateHash
      <3>2. LL1RAM.authenticator \in LL1ObservedAuthenticators
        <4>1. historyStateBinding \in HashType
          BY <1>3
        <4>2. UnforgeabilityInvariant
          BY <1>1
        <4>3. QED
          BY <3>1, <4>1, <4>2 DEF UnforgeabilityInvariant
      <3>3. QED
        BY <3>1, <3>2 DEF LL1HistoryStateBindingAuthenticated
    (***********************************************************************)
    (* Then, we can apply the InclusionInvariant to show that the output   *)
    (* from the service is in the set of observed outputs.                 *)
    (***********************************************************************)
    <2>6. QED
      <3>1. InclusionInvariant
        BY <1>1
      <3>2. QED
        BY <2>2, <2>3, <2>4, <2>5, <3>1
        DEF InclusionInvariant, sResult, privateState, historyStateBinding, stateHash
  (*************************************************************************)
  (* Since the element being uniond into the set is already in the set,    *)
  (* the set does not change.                                              *)
  (*************************************************************************)
  <1>6. QED
    BY <1>4, <1>5

----

(***************************************************************************)
(* The LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma   *)
(* states that the LL1RepeatOperation action does not change the set of    *)
(* history state bindings that have authenticators in the set              *)
(* LL1ObservedAuthenticators.  Together with the                           *)
(* LL1RepeatOperationUnchangedObseredOutputsLemma above, this is the       *)
(* essence of why LL1RepeatOperation does not cause any change to the      *)
(* refined high-level state.                                               *)
(***************************************************************************)

THEOREM LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma == \*`^\label{THEOREM-LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma}^'
    LL1TypeInvariant /\ UnforgeabilityInvariant /\ InclusionInvariant /\ LL1RepeatOperation =>
        \A historyStateBinding1 \in HashType :
            UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
  (*************************************************************************)
  (* We assume the antecedents.                                            *)
  (*************************************************************************)
  <1>1. HAVE LL1TypeInvariant /\ UnforgeabilityInvariant /\ InclusionInvariant /\ LL1RepeatOperation
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a new history *)
  (* state binding in HashType.                                            *)
  (*************************************************************************)
  <1>2. TAKE historyStateBinding1 \in HashType
  (*************************************************************************)
  (* Then, we pick some input for which LL1RepeatOperation is true.        *)
  (*************************************************************************)
  <1>3. PICK input \in LL1AvailableInputs : LL1RepeatOperation!(input)!1 
    BY <1>1 DEF LL1RepeatOperation
  (*************************************************************************)
  (* To simplify the writing of the proof, we re-state the definitions     *)
  (* from the LL1RepeatOperation action.                                   *)
  (*************************************************************************)
  <1> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
  <1> historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
  <1> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
  <1> sResult == Service(LL1RAM.publicState, privateState, input)
  <1> newPrivateStateEnc ==
          SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
  <1> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
  <1> newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
  <1> newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
  (*************************************************************************)
  (* We then assert the type safety of these definitions, with the help of *)
  (* the LL1RepeatOperationDefsTypeSafeLemma.                              *)
  (*************************************************************************)
  <1>4. /\  stateHash \in HashType
        /\  historyStateBinding \in HashType
        /\  privateState \in PrivateStateType
        /\  sResult \in ServiceResultType
        /\  sResult.newPublicState \in PublicStateType
        /\  sResult.newPrivateState \in PrivateStateType
        /\  sResult.output \in OutputType
        /\  newPrivateStateEnc \in PrivateStateEncType
        /\  newStateHash \in HashType
        /\  newHistoryStateBinding \in HashType
        /\  newAuthenticator \in MACType
    <2>1. input \in LL1AvailableInputs
      BY <1>3
    <2>2. LL1TypeInvariant
      BY <1>1
    <2>3. QED
      BY <2>1, <2>2, LL1RepeatOperationDefsTypeSafeLemma
  (*************************************************************************)
  (* We hide the definitions, so they don't overwhelm the prover.  We'll   *)
  (* pull them in as necessary below.                                      *)
  (*************************************************************************)
  <1> HIDE DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
          newStateHash, newHistoryStateBinding, newAuthenticator
  (*************************************************************************)
  (* From the definition of LL1RepeatOperation, we see the primed state of *)
  (* LL1ObservedAuthenticators is formed by unioning in the new            *)
  (* authenticator.                                                        *)
  (*************************************************************************)
  <1>5. LL1ObservedAuthenticators' =
            LL1ObservedAuthenticators \union {newAuthenticator}
    BY <1>3 DEF LL1RepeatOperation, newAuthenticator, newHistoryStateBinding,
                newStateHash, newPrivateStateEnc, sResult, privateState
  (*************************************************************************)
  (* One fact that will be useful in several places is that the symmetric  *)
  (* key in the NVRAM has not changed.                                     *)
  (*************************************************************************)
  <1>6. UNCHANGED LL1NVRAM.symmetricKey
    <2>1. UNCHANGED LL1NVRAM
      BY <1>3
    <2>2. QED
      BY <2>1
  (*************************************************************************)
  (* To show that LL1HistoryStateBindingAuthenticated predicate is         *)
  (* unchanged for all history state bindings, we first consider one       *)
  (* specific history state binding, namely the the new history state      *)
  (* binding defined in LL1RepeatOperation.                                *)
  (*************************************************************************)
  <1>7. CASE historyStateBinding1 = newHistoryStateBinding
    (***********************************************************************)
    (* First, we'll show that the new history state binding is             *)
    (* authenticated in the unprimed state.                                *)
    (***********************************************************************)
    <2>1. LL1HistoryStateBindingAuthenticated(newHistoryStateBinding)
      (*********************************************************************)
      (* Our strategy is to use the InclusionInvariant.  We first have to  *)
      (* show that all of the types are satisfied.                         *)
      (*********************************************************************)
      <3>2. LL1TypeInvariant
        BY <1>1
      <3>3. input \in InputType
        <4>1. input \in LL1AvailableInputs
          BY <1>3
        <4>2. LL1AvailableInputs \subseteq InputType
          BY <3>2 DEF LL1TypeInvariant
        <4>3. QED
          BY <4>1, <4>2
      <3>4. /\  LL1RAM.historySummary \in HashType
            /\  LL1RAM.publicState \in PublicStateType
            /\  LL1RAM.privateStateEnc \in PrivateStateEncType
        BY <3>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
      (*********************************************************************)
      (* We then have to show that the antecedents in the                  *)
      (* InclusionInvariant are satisfied.                                 *)
      (*********************************************************************)
      <3>5. LL1NVRAM.historySummary = Hash(LL1RAM.historySummary, input)
        BY <1>3
      <3>6. LL1HistoryStateBindingAuthenticated(historyStateBinding)
        (*******************************************************************)
        (* To show that the history state binding is authenticated, we     *)
        (* demonstrate that LL1RAM.authenticator is a sufficient witness   *)
        (* for the existential quantifier within the definition of         *)
        (* LL1HistoryStateBindingAuthenticated.                            *)
        (*******************************************************************)
        <4>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
          BY <1>3 DEF historyStateBinding, stateHash
        <4>2. LL1RAM.authenticator \in LL1ObservedAuthenticators
          <5>1. historyStateBinding \in HashType
            BY <1>4
          <5>2. UnforgeabilityInvariant
            BY <1>1
          <5>3. QED
            BY <4>1, <5>1, <5>2 DEF UnforgeabilityInvariant
        <4>3. QED
          BY <4>1, <4>2 DEF LL1HistoryStateBindingAuthenticated
      (*********************************************************************)
      (* Then, we can apply the InclusionInvariant to show that the new    *)
      (* history state binding is authenticated.                           *)
      (*********************************************************************)
      <3>7. QED
        <4>1. InclusionInvariant
          BY <1>1
        <4>2. QED
          BY <3>3, <3>4, <3>5, <3>6, <4>1
          DEF InclusionInvariant, newAuthenticator, newHistoryStateBinding,
              newStateHash, newPrivateStateEnc, sResult, privateState,
              historyStateBinding, stateHash
    (***********************************************************************)
    (* Next, we'll show that the new history state binding is              *)
    (* authenticated in the primed state.                                  *)
    (***********************************************************************)
    <2>2. LL1HistoryStateBindingAuthenticated(historyStateBinding1)'
      (*********************************************************************)
      (* By expanding the definition of                                    *)
      (* LL1HistoryStateBindingAuthenticated, it suffices to show that the *)
      (* new authenticator defined in LL1RepeatOperation (which we know to *)
      (* be in the primed set of observed authenticators) is a valid MAC   *)
      (* for the history state binding in the primed state.                *)
      (*********************************************************************)
      <3>1. SUFFICES ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding1, newAuthenticator)
        <4>1. newAuthenticator \in LL1ObservedAuthenticators'
          BY <1>5
        <4>2. QED
          BY <4>1 DEF LL1HistoryStateBindingAuthenticated
      (*********************************************************************)
      (* The new authenticator was genererated as a MAC of this history    *)
      (* state binding by LL1RepeatOperation in the unprimed state, and    *)
      (* the symmetric key in the NVRAM has not changed.                   *)
      (*********************************************************************)
      <3>2. newAuthenticator = GenerateMAC(LL1NVRAM.symmetricKey', historyStateBinding1)
        BY <1>6, <1>7 DEF newAuthenticator
      (*********************************************************************)
      (* We can thus use the MACComplete property to show that the         *)
      (* generated MAC validates appropriately.  To do this, we first need *)
      (* to prove some types.                                              *)
      (*********************************************************************)
      <3>3. LL1NVRAM.symmetricKey' \in SymmetricKeyType
        <4>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
          <5>1. LL1TypeInvariant
            BY <1>1
          <5>2. QED
            BY <5>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
        <4>2. QED
          BY <1>6, <4>1
      <3>4. historyStateBinding1 \in HashType
        BY <1>2
      (*********************************************************************)
      (* Then, we appeal to the MACComplete property in a straightforward  *)
      (* way.                                                              *)
      (*********************************************************************)
      <3>5. QED
        BY <3>2, <3>3, <3>4, MACComplete
    (***********************************************************************)
    (* Because the new history state binding is authenticated in both the  *)
    (* unprimed and primed states, the LL1HistoryStateBindingAuthenticated *)
    (* is unchanged for this history state binding.                        *)
    (***********************************************************************)
    <2>3. QED
      BY <1>7, <2>1, <2>2
  (*************************************************************************)
  (* We then consider every state binding that is not equal to the the new *)
  (* history state binding defined in LL1RepeatOperation.                  *)
  (*************************************************************************)
  <1>8. CASE historyStateBinding1 # newHistoryStateBinding
    (***********************************************************************)
    (* We'll subdivide these into two cases.  In the first case, we'll     *)
    (* consider the history state bindings that are authenticated in the   *)
    (* unprimed state, and we'll show that they continue to be             *)
    (* authenticated in the primed state.                                  *)
    (***********************************************************************)
    <2>1. CASE LL1HistoryStateBindingAuthenticated(historyStateBinding1) = TRUE
      <3>1. LL1HistoryStateBindingAuthenticated(historyStateBinding1)' = TRUE
        (*******************************************************************)
        (* By hypothesis, the history state binding is authenticated in    *)
        (* the unprimed state.  Thus, we can pick an authenticator in the  *)
        (* set of observed authenticators that is a valid MAC for this     *)
        (* history state binding.                                          *)
        (*******************************************************************)
        <4>1. PICK authenticator \in LL1ObservedAuthenticators :
                  ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding1, authenticator)
          BY <2>1 DEF LL1HistoryStateBindingAuthenticated
        (*******************************************************************)
        (* Because the symmetric key in the NVRAM has not changed, this    *)
        (* authenticator is also a valid MAC for this history state        *)
        (* binding in the primed state.                                    *)
        (*******************************************************************)
        <4>2. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding1, authenticator)
          BY <1>6, <4>1
        (*******************************************************************)
        (* Because the primed set of observed authenticators includes all  *)
        (* authenticators that were in the unprimed set, this              *)
        (* authenticator is also in the primed set of observed             *)
        (* authenticators.                                                 *)
        (*******************************************************************)
        <4>3. authenticator \in LL1ObservedAuthenticators'
          <5>1. authenticator \in LL1ObservedAuthenticators
            BY <4>1
          <5>2. LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
            BY <1>5
          <5>3. QED
            BY <5>1, <5>2
        (*******************************************************************)
        (* The previous two conditions are sufficient to establish that    *)
        (* the history state binding is authenticated in the primed state. *)
        (*******************************************************************)
        <4>4. QED
          BY <4>2, <4>3 DEF LL1HistoryStateBindingAuthenticated
      (*********************************************************************)
      (* Because the history state binding is authenticated in both the    *)
      (* unprimed and primed states, the                                   *)
      (* LL1HistoryStateBindingAuthenticated is unchanged for this history *)
      (* state binding.                                                    *)
      (*********************************************************************)
      <3>2. QED
        BY <2>1, <3>1
    (***********************************************************************)
    (* We'll subdivide these into two cases.  In the second case, we'll    *)
    (* consider the history state bindings that are unauthenticated in the *)
    (* unprimed state, and we'll show that they continue to be             *)
    (* unauthenticated in the primed state.                                *)
    (***********************************************************************)
    <2>2. CASE LL1HistoryStateBindingAuthenticated(historyStateBinding1) = FALSE
      <3>1. LL1HistoryStateBindingAuthenticated(historyStateBinding1)' = FALSE
        (*******************************************************************)
        (* To prove that the history state binding is not authenticated in *)
        (* the primed state, it suffices to show that none of the          *)
        (* authenticators in the primed set of observed authenticators is  *)
        (* a valid MAC for the history state binding.                      *)
        (*******************************************************************)
        <4>1. SUFFICES \A authenticator \in LL1ObservedAuthenticators' :
                  ~ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding1, authenticator)
          BY DEF LL1HistoryStateBindingAuthenticated
        (*******************************************************************)
        (* To prove the universally quantified expression, we take a new   *)
        (* authenticator in the primed set of observed authenticators.     *)
        (*******************************************************************)
        <4>2. TAKE authenticator \in LL1ObservedAuthenticators'
        (*******************************************************************)
        (* We'll subdivide this into two cases.  First, we consider the    *)
        (* case in which the authenticator is in the unprimed set of       *)
        (* authenticators.  In this case, because the authenticator failed *)
        (* to authenticate the history state binding in the unprimed       *)
        (* state, and the symmetric key has not changed, it immediately    *)
        (* follows that the authenticator will not authenticate the        *)
        (* history state binding in the primed state.                      *)
        (*******************************************************************)
        <4>3. CASE authenticator \in LL1ObservedAuthenticators
          BY <1>6, <2>2, <4>3 DEF LL1HistoryStateBindingAuthenticated
        (*******************************************************************)
        (* In the second case, we consider the new authenticator defined   *)
        (* in LL1RepeatOperation.                                          *)
        (*******************************************************************)
        <4>4. CASE authenticator = newAuthenticator
          (*****************************************************************)
          (* We'll use proof by contradiction.  Assume that the new        *)
          (* authenticator is a valid MAC for the history state binding.   *)
          (*****************************************************************)
          <5>1. SUFFICES
                ASSUME ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding1, authenticator)
                PROVE FALSE
            OBVIOUS
          (*****************************************************************)
          (* By the collision resistance of MACs, it must be the case the  *)
          (* history state binding is equal to the new history state binding *)
          (* defined in LL1RepeatOperation.                                *)
          (*****************************************************************)
          <5>2. historyStateBinding1 = newHistoryStateBinding
            <6>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
              <7>1. LL1TypeInvariant
                BY <1>1
              <7>2. QED
                BY <7>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
            <6>2. LL1NVRAM.symmetricKey' \in SymmetricKeyType
              BY <1>6, <6>1
            <6>3. historyStateBinding1 \in HashType
              BY <1>2
            <6>4. newHistoryStateBinding \in HashType
              BY <1>4
            <6>5. QED
              BY <4>4, <5>1, <6>1, <6>2, <6>3, <6>4, MACCollisionResistant DEF newAuthenticator
          (*****************************************************************)
          (* But we are working within a case in which the history state   *)
          (* binding is not equal to the new history state binding defined *)
          (* in LL1RepeatOperation.  Thus, we have a contradiction.        *)
          (*****************************************************************)
          <5>3. QED
            BY <1>8, <5>2
        (*******************************************************************)
        (* We've considered authenticators in the unprimed set of          *)
        (* authenticators, and we've considered the new authenticator      *)
        (* defined in LL1RepeatOperation.  Because the primed set of       *)
        (* authenticators is the union of these two, we have exhausted the *)
        (* cases.                                                          *)
        (*******************************************************************)
        <4>5. QED
          BY <1>5, <4>3, <4>4
      (*********************************************************************)
      (* Because the history state binding is unauthenticated in both the  *)
      (* unprimed and primed states, the                                   *)
      (* LL1HistoryStateBindingAuthenticated is unchanged for this history *)
      (* state binding.                                                    *)
      (*********************************************************************)
      <3>2. QED
        BY <2>2, <3>1
    (***********************************************************************)
    (* By proving that LL1HistoryStateBindingAuthenticated is a boolean    *)
    (* predicate, it is immediately clear that the two cases of true and   *)
    (* false are exhaustive for this predicate.                            *)
    (***********************************************************************)
    <2>3. LL1HistoryStateBindingAuthenticated(historyStateBinding1) \in BOOLEAN
      BY DEF LL1HistoryStateBindingAuthenticated
    <2>4. QED
      BY <2>1, <2>2, <2>3
  (*************************************************************************)
  (* Because the conclusion holds for (1) the new history state binding    *)
  (* defined in LL1RepeatOperation and (2) every other state binding, the  *)
  (* conclusion holds for all state bindings.                              *)
  (*************************************************************************)
  <1>9. QED
    BY <1>7, <1>8

----

(***************************************************************************)
(* The LL1RAMUnforgeabilityUnchangedLemma states that, if the symmetric    *)
(* key in the NVRAM does not change and the set of observed authenticators *)
(* does not change, then the RAM's portion of the                          *)
(* ExtendedUnforgeabilityInvariant inductively holds when the RAM's primed *)
(* value is taken from the RAM or the disk.                                *)
(***************************************************************************)

THEOREM LL1RAMUnforgeabilityUnchangedLemma == \*`^\label{THEOREM-LL1RAMUnforgeabilityUnchangedLemma}^'
      ( /\  ExtendedUnforgeabilityInvariant
        /\  LL1TypeInvariant
        /\  LL1TypeInvariant'
        /\  LL1RAM' \in {LL1RAM, LL1Disk}
        /\  LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
        /\  UNCHANGED LL1NVRAM.symmetricKey )
    =>
        \A historyStateBinding \in HashType :
            ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator') =>
                LL1RAM.authenticator' \in LL1ObservedAuthenticators'
  (*************************************************************************)
  (* We begin by assuming the antecedent.                                  *)
  (*************************************************************************)
  <1>1. HAVE /\  ExtendedUnforgeabilityInvariant
             /\  LL1TypeInvariant
             /\  LL1TypeInvariant'
             /\  LL1RAM' \in {LL1RAM, LL1Disk}
             /\  LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
             /\  UNCHANGED LL1NVRAM.symmetricKey
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a new         *)
  (* historyStateBinding in the HashType.                                  *)
  (*************************************************************************)
  <1>2. TAKE historyStateBinding \in HashType
  (*************************************************************************)
  (* We then assume the antecedent in the nested implication.              *)
  (*************************************************************************)
  <1>3. HAVE ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1RAM.authenticator')
  (*************************************************************************)
  (* The LL1RAM's primed state is taken from either the LL1RAM or the      *)
  (* disk.                                                                 *)
  (*************************************************************************)
  <1>4. LL1RAM' \in {LL1RAM, LL1Disk}
    BY <1>1
  (*************************************************************************)
  (* Case 1: the LL1RAM's primed state comes from the LL1RAM.  There are   *)
  (* three basic steps.                                                    *)
  (*************************************************************************)
  <1>5. CASE UNCHANGED LL1RAM
    (***********************************************************************)
    (* First, since we take the antecedent in the nested implication and   *)
    (* swap out unprimed variables for primed variables, since the         *)
    (* symmetric key and authenticator have not changed.                   *)
    (***********************************************************************)
    <2>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
      <3>1. UNCHANGED LL1NVRAM.symmetricKey
        BY <1>1
      <3>2. UNCHANGED LL1RAM.authenticator
        BY <1>5
      <3>3. QED
        BY <1>3, <3>1, <3>2
    (***********************************************************************)
    (* Second, using the ExtendedUnforgeabilityInvariant, we show that the *)
    (* authenticator was in the unprimed set of observed authenticators.   *)
    (***********************************************************************)
    <2>2. LL1RAM.authenticator \in LL1ObservedAuthenticators
      <3>1. ExtendedUnforgeabilityInvariant
        BY <1>1
      <3>2. QED
        BY <1>2, <2>1, <3>1 DEF ExtendedUnforgeabilityInvariant
    (***********************************************************************)
    (* Third, we show that the authenticator is also in the primed set of  *)
    (* observed authenticators, since the symmetric key has not changed    *)
    (* and set of observed authenticators includes every element in the    *)
    (* primed state that it included in the unprimed state.                *)
    (***********************************************************************)
    <2>3. QED
      <3>1. UNCHANGED LL1RAM.authenticator
        BY <1>5
      <3>2. LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
        BY <1>1
      <3>3. QED
        BY <2>2, <3>1, <3>2
  (*************************************************************************)
  (* Case 2: the RAM's primed state comes from the disk.  The proof is     *)
  (* straightforward.                                                      *)
  (*************************************************************************)
  <1>6. CASE LL1RAM' = LL1Disk
    (***********************************************************************)
    (* First, since we take the antecedent in the nested implication and   *)
    (* make two changes: (1) swap out unprimed variables for primed        *)
    (* variables and (2) replace LL1RAM with LL1Disk, since the primed     *)
    (* state of the RAM comes from the unprimed state of the disk.         *)
    (***********************************************************************)
    <2>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1Disk.authenticator)
      <3>1. UNCHANGED LL1NVRAM.symmetricKey
        BY <1>1
      <3>2. LL1RAM.authenticator' = LL1Disk.authenticator
        BY <1>6
      <3>3. QED
        BY <1>3, <3>1, <3>2
    (***********************************************************************)
    (* Second, using the ExtendedUnforgeabilityInvariant, we show that the *)
    (* authenticator was in the unprimed set of observed authenticators.   *)
    (***********************************************************************)
    <2>2. LL1Disk.authenticator \in LL1ObservedAuthenticators
      <3>1. ExtendedUnforgeabilityInvariant
        BY <1>1
      <3>2. QED
        BY <1>2, <2>1, <3>1 DEF ExtendedUnforgeabilityInvariant
    (***********************************************************************)
    (* Third, we show that the authenticator is also in the primed set of  *)
    (* observed authenticators, since the symmetric key has not changed    *)
    (* and set of observed authenticators includes every element in the    *)
    (* primed state that it included in the unprimed state.                *)
    (***********************************************************************)
    <2>3. QED
      <3>1. LL1RAM.authenticator' = LL1Disk.authenticator
        BY <1>6
      <3>2. LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
        BY <1>1
      <3>3. QED
        BY <2>2, <3>1, <3>2
  (*************************************************************************)
  (* The theorem is true by exhaustive case analysis.                      *)
  (*************************************************************************)
  <1>7. QED
    BY <1>4, <1>5, <1>6

(***************************************************************************)
(* The LL1DiskUnforgeabilityUnchangedLemma states that, if the symmetric   *)
(* key in the NVRAM does not change and the set of observed authenticators *)
(* does not change, then the disk's portion of the                         *)
(* ExtendedUnforgeabilityInvariant inductively holds when the disk's       *)
(* primed value is taken from the RAM or the disk.                         *)
(***************************************************************************)

THEOREM LL1DiskUnforgeabilityUnchangedLemma == \*`^\label{THEOREM-LL1DiskUnforgeabilityUnchangedLemma}^'
      ( /\  ExtendedUnforgeabilityInvariant
        /\  LL1TypeInvariant
        /\  LL1TypeInvariant'
        /\  LL1Disk' \in {LL1RAM, LL1Disk}
        /\  LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
        /\  UNCHANGED LL1NVRAM.symmetricKey )
    =>
        \A historyStateBinding \in HashType :
            ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1Disk.authenticator') =>
                LL1Disk.authenticator' \in LL1ObservedAuthenticators'
  (*************************************************************************)
  (* We begin by assuming the antecedent.                                  *)
  (*************************************************************************)
  <1>1. HAVE /\  ExtendedUnforgeabilityInvariant
             /\  LL1TypeInvariant
             /\  LL1TypeInvariant'
             /\  LL1Disk' \in {LL1RAM, LL1Disk}
             /\  LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
             /\  UNCHANGED LL1NVRAM.symmetricKey
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a new         *)
  (* historyStateBinding in the HashType.                                  *)
  (*************************************************************************)
  <1>2. TAKE historyStateBinding \in HashType
  (*************************************************************************)
  (* We then assume the antecedent in the nested implication.              *)
  (*************************************************************************)
  <1>3. HAVE ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, LL1Disk.authenticator')
  (*************************************************************************)
  (* The disk's primed state is taken from either the RAM or the disk.     *)
  (*************************************************************************)
  <1>4. LL1Disk' \in {LL1RAM, LL1Disk}
    BY <1>1
  (*************************************************************************)
  (* Case 1: the disk's primed state comes from the disk.  There are three *)
  (* basic steps.                                                          *)
  (*************************************************************************)
  <1>5. CASE UNCHANGED LL1Disk
    (***********************************************************************)
    (* First, since we take the antecedent in the nested implication and   *)
    (* swap out unprimed variables for primed variables, since the         *)
    (* symmetric key and authenticator have not changed.                   *)
    (***********************************************************************)
    <2>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1Disk.authenticator)
      <3>1. UNCHANGED LL1NVRAM.symmetricKey
        BY <1>1
      <3>2. UNCHANGED LL1Disk.authenticator
        BY <1>5
      <3>3. QED
        BY <1>3, <3>1, <3>2
    (***********************************************************************)
    (* Second, using the ExtendedUnforgeabilityInvariant, we show that the *)
    (* authenticator was in the unprimed set of observed authenticators.   *)
    (***********************************************************************)
    <2>2. LL1Disk.authenticator \in LL1ObservedAuthenticators
      <3>1. ExtendedUnforgeabilityInvariant
        BY <1>1
      <3>2. QED
        BY <1>2, <2>1, <3>1 DEF ExtendedUnforgeabilityInvariant
    (***********************************************************************)
    (* Third, we show that the authenticator is also in the primed set of  *)
    (* observed authenticators, since the symmetric key has not changed    *)
    (* and set of observed authenticators includes every element in the    *)
    (* primed state that it included in the unprimed state.                *)
    (***********************************************************************)
    <2>3. QED
      <3>1. UNCHANGED LL1Disk.authenticator
        BY <1>5
      <3>2. LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
        BY <1>1
      <3>3. QED
        BY <2>2, <3>1, <3>2
  (*************************************************************************)
  (* Case 2: the disk's primed state comes from the RAM.  The proof is     *)
  (* straightforward.                                                      *)
  (*************************************************************************)
  <1>6. CASE LL1Disk' = LL1RAM
    (***********************************************************************)
    (* First, since we take the antecedent in the nested implication and   *)
    (* make two changes: (1) swap out unprimed variables for primed        *)
    (* variables and (2) replace LL1Disk with LL1RAM, since the primed     *)
    (* state of the disk comes from the unprimed state of the RAM.         *)
    (***********************************************************************)
    <2>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
      <3>1. UNCHANGED LL1NVRAM.symmetricKey
        BY <1>1
      <3>2. LL1Disk.authenticator' = LL1RAM.authenticator
        BY <1>6
      <3>3. QED
        BY <1>3, <3>1, <3>2
    (***********************************************************************)
    (* Second, using the ExtendedUnforgeabilityInvariant, we show that the *)
    (* authenticator was in the unprimed set of observed authenticators.   *)
    (***********************************************************************)
    <2>2. LL1RAM.authenticator \in LL1ObservedAuthenticators
      <3>1. ExtendedUnforgeabilityInvariant
        BY <1>1
      <3>2. QED
        BY <1>2, <2>1, <3>1 DEF ExtendedUnforgeabilityInvariant
    (***********************************************************************)
    (* Third, we show that the authenticator is also in the primed set of  *)
    (* observed authenticators, since the symmetric key has not changed    *)
    (* and set of observed authenticators includes every element in the    *)
    (* primed state that it included in the unprimed state.                *)
    (***********************************************************************)
    <2>3. QED
      <3>1. LL1Disk.authenticator' = LL1RAM.authenticator
        BY <1>6
      <3>2. LL1ObservedAuthenticators \subseteq LL1ObservedAuthenticators'
        BY <1>1
      <3>3. QED
        BY <2>2, <3>1, <3>2
  (*************************************************************************)
  (* The theorem is true by exhaustive case analysis.                      *)
  (*************************************************************************)
  <1>7. QED
    BY <1>4, <1>5, <1>6

----

(***************************************************************************)
(* The InclusionUnchangedLemma states that, if there are no changes to the *)
(* NVRAM, to the set of observed outputs, or to the authentication status  *)
(* of any history state binding, then the InclusionInvariant holds         *)
(* inductively from the unprimed state to the primed state.                *)
(***************************************************************************)

THEOREM InclusionUnchangedLemma == \*`^\label{THEOREM-InclusionUnchangedLemma}^'
      ( /\  InclusionInvariant
        /\  LL1TypeInvariant
        /\  LL1TypeInvariant'
        /\  UNCHANGED <<LL1NVRAM, LL1ObservedOutputs>>
        /\  \A historyStateBinding1 \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1) )
    =>
        InclusionInvariant'
  (*************************************************************************)
  (* We begin by assuming the antecedent.                                  *)
  (*************************************************************************)
  <1>1. HAVE /\  InclusionInvariant
             /\  LL1TypeInvariant
             /\  LL1TypeInvariant'
             /\  UNCHANGED <<LL1NVRAM, LL1ObservedOutputs>>
             /\  \A historyStateBinding1 \in HashType :
                     UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a new set of  *)
  (* variables in the appropriate types.  For the TAKE step to be          *)
  (* meaningful to the prover, first we have to tell the prover to expand  *)
  (* the definition of InclusionInvariant, so it will see the universally  *)
  (* quantified expression therein.                                        *)
  (*************************************************************************)
  <1> USE DEF InclusionInvariant
  <1>2. TAKE  input \in InputType,
              historySummary \in HashType,
              publicState \in PublicStateType,
              privateStateEnc \in PrivateStateEncType
  (*************************************************************************)
  (* To simplify the writing of the proof, we re-state the definitions     *)
  (* from the InclusionInvariant.                                          *)
  (*************************************************************************)
  <1> stateHash == Hash(publicState, privateStateEnc)
  <1> historyStateBinding == Hash(historySummary, stateHash)
  <1> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, privateStateEnc)
  <1> sResult == Service(publicState, privateState, input)
  <1> newPrivateStateEnc ==
          SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
  <1> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
  <1> newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
  (*************************************************************************)
  (* We then assert the type safety of these definitions, with the help of *)
  (* the InclusionInvariantDefsTypeSafeLemma.                              *)
  (*************************************************************************)
  <1>3. /\  stateHash \in HashType
        /\  historyStateBinding \in HashType
        /\  privateState \in PrivateStateType
        /\  sResult \in ServiceResultType
        /\  sResult.newPublicState \in PublicStateType
        /\  sResult.newPrivateState \in PrivateStateType
        /\  sResult.output \in OutputType
        /\  newPrivateStateEnc \in PrivateStateEncType
        /\  newStateHash \in HashType
        /\  newHistoryStateBinding \in HashType
    <2>1. LL1TypeInvariant
      BY <1>1
    <2>2. QED
      BY <1>2, <2>1, InclusionInvariantDefsTypeSafeLemma
  (*************************************************************************)
  (* The InclusionInvariant states an implication.  To prove this, it      *)
  (* suffices to assume the antecedent and prove the consequent.           *)
  (*************************************************************************)
  <1>4. SUFFICES
        ASSUME
            /\  LL1NVRAM.historySummary' = Hash(historySummary, input)
            /\  LL1HistoryStateBindingAuthenticated(historyStateBinding)'
        PROVE
            /\  sResult.output' \in LL1ObservedOutputs'
            /\  LL1HistoryStateBindingAuthenticated(newHistoryStateBinding)'
    OBVIOUS
  (*************************************************************************)
  (* We hide the definitions, so they don't overwhelm the prover.  We'll   *)
  (* pull them in as necessary below.                                      *)
  (*************************************************************************)
  <1> HIDE DEF InclusionInvariant
  <1> HIDE DEF stateHash, historyStateBinding, privateState, sResult,
          newPrivateStateEnc, newStateHash, newHistoryStateBinding
  (*************************************************************************)
  (* We will prove each of the conjuncts separately.  Following is the     *)
  (* proof of the first conjunct.                                          *)
  (*                                                                       *)
  (* To prove that the primed output of the service is in the primed set   *)
  (* of observed outputs, we prove three things: (1) Before the action,    *)
  (* the output of the service is in the set of observed outputs.  (2) The *)
  (* output of the service does not change.  (3) The set of observed       *)
  (* outputs does not change.                                              *)
  (*************************************************************************)
  <1>5. sResult.output' \in LL1ObservedOutputs'
    (***********************************************************************)
    (* Step 1: Before the action, the output of the service is in the set  *)
    (* of observed outputs.  This follows because the InclusionInvariant   *)
    (* is true in the unprimed state.                                      *)
    (***********************************************************************)
    <2>1. sResult.output \in LL1ObservedOutputs
      (*********************************************************************)
      (* We prove the two conjuncts in the antecedent of the               *)
      (* InclusionInvariant.  Each follows as a straightforward            *)
      (* consequence of the fact that LL1NVRAM and                         *)
      (* LL1ObservedAuthenticators have not changed.                       *)
      (*********************************************************************)
      <3>1. LL1NVRAM.historySummary = Hash(historySummary, input)
        <4>1. LL1NVRAM.historySummary' = Hash(historySummary, input)
          BY <1>4
        <4>2. UNCHANGED LL1NVRAM.historySummary
          <5>1. UNCHANGED LL1NVRAM
            BY <1>1
          <5>2. QED
            BY <5>1
        <4>3. QED
          BY <4>1, <4>2
      <3>2. LL1HistoryStateBindingAuthenticated(historyStateBinding)
        (*******************************************************************)
        (* The history state binding is authenticated in the primed state, *)
        (* by hypothesis.                                                  *)
        (*******************************************************************)
        <4>1. LL1HistoryStateBindingAuthenticated(historyStateBinding)'
          BY <1>4
        (*******************************************************************)
        (* The authentication status of the history state binding has not  *)
        (* changed, because this status has not changed for any history    *)
        (* state binding in HashType.                                      *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          (*****************************************************************)
          (* We have to show that the history state binding has the        *)
          (* appropriate type.                                             *)
          (*****************************************************************)
          <5>1. historyStateBinding \in HashType
            BY <1>3
          (*****************************************************************)
          (* The authentication status of all input state bindings has not *)
          (* changed, as assumed by the lemma.                             *)
          (*****************************************************************)
          <5>2. \A historyStateBinding1 \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
            BY <1>1
          (*****************************************************************)
          (* The conclusion follows directly.                              *)
          (*****************************************************************)
          <5>3. QED
            BY <5>1, <5>2
        (*******************************************************************)
        (* Since the history state binding is authenticated in the primed  *)
        (* state, and since its authentication status has not changed, it  *)
        (* is also authenticated in the unprimed state.                    *)
        (*******************************************************************)
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We can then use the InclusionInvariant to prove that the output   *)
      (* of the service is in the set of observed outputs in the unprimed  *)
      (* state.                                                            *)
      (*********************************************************************)
      <3>3. InclusionInvariant
        BY <1>1
      <3>4. QED
        BY <1>2, <3>1, <3>2, <3>3
        DEF InclusionInvariant, sResult, privateState, historyStateBinding, stateHash
    (***********************************************************************)
    (* Step 2: The output of the service does not change.                  *)
    (***********************************************************************)
    <2>2. UNCHANGED sResult.output
      <3>1. UNCHANGED privateState
        <4>1. UNCHANGED LL1NVRAM.symmetricKey
          <5>1. UNCHANGED LL1NVRAM
            BY <1>1
          <5>2. QED
            BY <5>1
        <4>2. QED
          BY <4>1 DEF privateState
      <3>2. UNCHANGED sResult
        BY <3>1 DEF sResult
      <3>3. QED
        BY <3>2
    (***********************************************************************)
    (* Step 3: The set of observed outputs does not change.                *)
    (***********************************************************************)
    <2>3. UNCHANGED LL1ObservedOutputs
      BY <1>1
    <2>4. QED
      BY <2>1, <2>2, <2>3
  (*************************************************************************)
  (* Following is the proof of the second conjunct.                        *)
  (*                                                                       *)
  (* To prove that the new history state binding is authenticated in the   *)
  (* primed state, we prove that the new history state binding does not    *)
  (* change, and we prove that the new history state binding was           *)
  (* authenticated in the unprimed state.  Since, by assumption of the     *)
  (* lemma, the authentication status of any history state binding does    *)
  (* not change, the new history state binding is authenticated in the     *)
  (* primed state.                                                         *)
  (*************************************************************************)
  <1>6. LL1HistoryStateBindingAuthenticated(newHistoryStateBinding)'
    (***********************************************************************)
    (* One fact we'll need several times is that the symmetric key in the  *)
    (* NVRAM has not changed, so we'll prove this once up front.           *)
    (***********************************************************************)
    <2>1. UNCHANGED LL1NVRAM.symmetricKey
      <3>1. UNCHANGED LL1NVRAM
        BY <1>1
      <3>2. QED
        BY <3>1
    (***********************************************************************)
    (* In the unprimed state, the new history state binding was            *)
    (* authenticated.  This follows because the InclusionInvariant is true *)
    (* in the unprimed state.                                              *)
    (***********************************************************************)
    <2>2. LL1HistoryStateBindingAuthenticated(newHistoryStateBinding)
      (*********************************************************************)
      (* We prove the two conjuncts in the antecedent of the               *)
      (* InclusionInvariant.  The first conjunct follows as a              *)
      (* straightforward consequence of the fact that NVRAM has not        *)
      (* changed.                                                          *)
      (*********************************************************************)
      <3>1. LL1NVRAM.historySummary = Hash(historySummary, input)
        <4>1. LL1NVRAM.historySummary' = Hash(historySummary, input)
          BY <1>4
        <4>2. UNCHANGED LL1NVRAM.historySummary
          <5>1. UNCHANGED LL1NVRAM
            BY <1>1
          <5>2. QED
            BY <5>1
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* Proving the second conjuct is more involved.  We'll show that the *)
      (* history state binding is authenticated in the primed state and    *)
      (* that its authentication status has not changed                    *)
      (*********************************************************************)
      <3>2. LL1HistoryStateBindingAuthenticated(historyStateBinding)
        (*******************************************************************)
        (* The history state binding is authenticated in the primed state, *)
        (* by hypothesis.                                                  *)
        (*******************************************************************)
        <4>1. LL1HistoryStateBindingAuthenticated(historyStateBinding)'
          BY <1>4
        (*******************************************************************)
        (* The authentication status of the history state binding has not  *)
        (* changed, because this status has not changed for any history    *)
        (* state binding in HashType.                                      *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          (*****************************************************************)
          (* We have to show that the history state binding has the        *)
          (* appropriate type.                                             *)
          (*****************************************************************)
          <5>1. historyStateBinding \in HashType
            BY <1>3
          (*****************************************************************)
          (* The authentication status of all input state bindings has not *)
          (* changed, as assumed by the lemma.                             *)
          (*****************************************************************)
          <5>2. \A historyStateBinding1 \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
            BY <1>1
          (*****************************************************************)
          (* The conclusion follows directly.                              *)
          (*****************************************************************)
          <5>3. QED
            BY <5>1, <5>2
        (*******************************************************************)
        (* Since the history state binding is authenticated in the primed  *)
        (* state, and since its authentication status has not changed, it  *)
        (* is also authenticated in the unprimed state.                    *)
        (*******************************************************************)
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We can then use the InclusionInvariant to prove that the new      *)
      (* history state binding was authenticated in the unprimed state.    *)
      (*********************************************************************)
      <3>3. InclusionInvariant
        BY <1>1
      <3>4. QED
        BY <1>2, <3>1, <3>2, <3>3
        DEF InclusionInvariant, newHistoryStateBinding, newStateHash, newPrivateStateEnc,
            sResult, privateState, historyStateBinding, stateHash
    <2>3. UNCHANGED LL1HistoryStateBindingAuthenticated(newHistoryStateBinding)
    (***********************************************************************)
    (* The new history state binding has not changed.                      *)
    (***********************************************************************)
      <3>1. UNCHANGED newHistoryStateBinding
        <4>1. UNCHANGED LL1NVRAM.historySummary
          <5>1. UNCHANGED LL1NVRAM
            BY <1>1
          <5>2. QED
            BY <5>1
        <4>2. UNCHANGED newStateHash
          <5>1. UNCHANGED sResult
            <6>1. UNCHANGED privateState
              BY <2>1 DEF privateState
            <6>2. QED
              BY <6>1 DEF sResult
          <5>2. UNCHANGED sResult.newPublicState
            BY <5>1
          <5>3. UNCHANGED newPrivateStateEnc
            <6>1. UNCHANGED sResult.newPrivateState
              BY <5>1
            <6>2. QED
              BY <2>1, <6>1 DEF newPrivateStateEnc
          <5>4. QED
            BY <5>2, <5>3 DEF newStateHash
        <4>3. QED
          BY <4>1, <4>2 DEF newHistoryStateBinding
      (*********************************************************************)
      (* The new history state binding has the appropriate type.           *)
      (*********************************************************************)
      <3>2. newHistoryStateBinding \in HashType
        BY <1>3
      (*********************************************************************)
      (* The authentication status of all input state bindings has not     *)
      (* changed, as assumed by the lemma.                                 *)
      (*********************************************************************)
      <3>3. \A historyStateBinding1 \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
        BY <1>1
      <3>4. QED
        BY <3>1, <3>2, <3>3
    <2>4. QED
      BY <2>2, <2>3
  (*************************************************************************)
  (* Each of the conjuncts is true, so the conjunction is true.            *)
  (*************************************************************************)
  <1>7. QED
    BY <1>5, <1>6

----

(***************************************************************************)
(* The CardinalityUnchangedLemma states that, if there are no changes to   *)
(* the NVRAM or to the authentication status of any history state binding, *)
(* then the CardinalityInvariant holds inductively from the unprimed state *)
(* to the primed state.                                                    *)
(***************************************************************************)

THEOREM CardinalityUnchangedLemma == \*`^\label{THEOREM-CardinalityUnchangedLemma}^'
      ( /\  LL1TypeInvariant
        /\  CardinalityInvariant
        /\  UNCHANGED LL1NVRAM
        /\  \A historyStateBinding1 \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1) )
    =>
        CardinalityInvariant'
  (*************************************************************************)
  (* We begin by assuming the antecedent.                                  *)
  (*************************************************************************)
  <1>1. HAVE /\  LL1TypeInvariant
             /\  CardinalityInvariant
             /\  UNCHANGED LL1NVRAM
             /\  \A historyStateBinding \in HashType :
                     UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a new set of  *)
  (* variables in the appropriate types.  For the TAKE step to be          *)
  (* meaningful to the prover, first we have to tell the prover to expand  *)
  (* the definition of CardinalityInvariant, so it will see the            *)
  (* universally quantified expression therein.                            *)
  (*************************************************************************)
  <1> USE DEF CardinalityInvariant
  <1>2. TAKE historySummary \in HashType, stateHash \in HashType
  (*************************************************************************)
  (* To simplify the writing of the proof, we re-state the definition from *)
  (* the CardinalityInvariant.                                             *)
  (*************************************************************************)
  <1> historyStateBinding == Hash(historySummary, stateHash)
  (*************************************************************************)
  (* We then assert the type safety of these definitions, with the help of *)
  (* the CardinalityInvariantDefsTypeSafeLemma.                            *)
  (*************************************************************************)
  <1>3. historyStateBinding \in HashType
      BY <1>2, CardinalityInvariantDefsTypeSafeLemma
  (*************************************************************************)
  (* The CardinalityInvariant states an implication.  To prove this, it    *)
  (* suffices to assume the antecedent and prove the consequent.           *)
  (*************************************************************************)
  <1>4. SUFFICES
        ASSUME
            /\  LL1NVRAMHistorySummaryUncorrupted'
            /\  LL1HistoryStateBindingAuthenticated(historyStateBinding)'
        PROVE
            HashCardinality(historySummary) <= HashCardinality(LL1NVRAM.historySummary')
    OBVIOUS
  (*************************************************************************)
  (* We hide the definitions, so they don't overwhelm the prover.  We'll   *)
  (* pull them in as necessary below.                                      *)
  (*************************************************************************)
  <1> HIDE DEF CardinalityInvariant
  <1> HIDE DEF historyStateBinding
  (*************************************************************************)
  (* To prove the inequality in the primed state, we prove two things: (1) *)
  (* Before the action, the inequality held.  (2) The history summary in   *)
  (* the NVRAM does not change.                                            *)
  (*                                                                       *)
  (* The inequality held before the action because the                     *)
  (* CardinalityInvariant was true in the unprimed state.                  *)
  (*************************************************************************)
  <1>5. HashCardinality(historySummary) <= HashCardinality(LL1NVRAM.historySummary)
    (***********************************************************************)
    (* We'll prove the two antecedents of the CardinalityInvariant in the  *)
    (* unprimed state.  First, we'll prove that the                        *)
    (* LL1NVRAMHistorySummaryUncorrupted predicate is true.                *)
    (***********************************************************************)
    <2>1. LL1NVRAMHistorySummaryUncorrupted
      (*********************************************************************)
      (* The LL1NVRAMHistorySummaryUncorrupted predicate is true in the    *)
      (* primed state, by hypothesis.                                      *)
      (*********************************************************************)
      <3>1. LL1NVRAMHistorySummaryUncorrupted'
        BY <1>4
      (*********************************************************************)
      (* The LL1NVRAMHistorySummaryUncorrupted predicate has not changed,  *)
      (* thanks to the LL1NVRAMHistorySummaryUncorruptedUnchangedLemma.    *)
      (*********************************************************************)
      <3>2. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
        BY <1>1, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
      (*********************************************************************)
      (* Since the LL1NVRAMHistorySummaryUncorrupted predicate is true in  *)
      (* the primed state, and since the LL1NVRAMHistorySummaryUncorrupted *)
      (* predicate has not changed, the predicate is also true in the      *)
      (* unprimed state.                                                   *)
      (*********************************************************************)
      <3>3. QED
        BY <3>1, <3>2
    (***********************************************************************)
    (* Then, we'll prove that the history state binding is authenticated   *)
    (* in the unprimed state.  This is the second antecedent of the        *)
    (* CardinalityInvariant.                                               *)
    (***********************************************************************)
    <2>2. LL1HistoryStateBindingAuthenticated(historyStateBinding)
      (*********************************************************************)
      (* The history state binding is authenticated in the primed state,   *)
      (* by hypothesis.                                                    *)
      (*********************************************************************)
      <3>1. LL1HistoryStateBindingAuthenticated(historyStateBinding)'
        BY <1>4
      (*********************************************************************)
      (* The authentication status of the history state binding has not    *)
      (* changed, because this status has not changed for any history      *)
      (* state binding in HashType.                                        *)
      (*********************************************************************)
      <3>2. UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
        (*******************************************************************)
        (* We have to show that the history state binding has the          *)
        (* appropriate type.                                               *)
        (*******************************************************************)
        <4>1. historyStateBinding \in HashType
          BY <1>3
        (*******************************************************************)
        (* The authentication status of all input state bindings has not   *)
        (* changed, as assumed by the lemma.                               *)
        (*******************************************************************)
        <4>2. \A historyStateBinding1 \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
          BY <1>1
        (*******************************************************************)
        (* The conclusion follows directly.                                *)
        (*******************************************************************)
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* Since the history state binding is authenticated in the primed    *)
      (* state, and since its authentication status has not changed, it is *)
      (* also authenticated in the unprimed state.                         *)
      (*********************************************************************)
      <3>3. QED
        BY <3>1, <3>2
    (***********************************************************************)
    (* We can then use the CardinalityInvariant to prove that the          *)
    (* inquality held in the unprimed state.                               *)
    (***********************************************************************)
    <2>3. CardinalityInvariant
      BY <1>1
    <2>4. QED
      BY <1>2, <2>1, <2>2, <2>3
      DEF CardinalityInvariant, historyStateBinding
  (*************************************************************************)
  (* Step 2: The history summary in the NVRAM does not change.             *)
  (*************************************************************************)
  <1>6. UNCHANGED LL1NVRAM.historySummary
    <2>1. UNCHANGED LL1NVRAM
      BY <1>1
    <2>2. QED
      BY <2>1
  <1>7. QED
    BY <1>5, <1>6

----

(***************************************************************************)
(* The UniquenessUnchangedLemma states that, if there are no changes to    *)
(* the NVRAM or to the authentication status of any history state binding, *)
(* then the UniquenessInvariant holds inductively from the unprimed state  *)
(* to the primed state.                                                    *)
(***************************************************************************)

THEOREM UniquenessUnchangedLemma == \*`^\label{THEOREM-UniquenessUnchangedLemma}^'
      ( /\  LL1TypeInvariant
        /\  UniquenessInvariant
        /\  UNCHANGED LL1NVRAM
        /\  \A historyStateBinding \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding) )
    =>
        UniquenessInvariant'
  (*************************************************************************)
  (* We begin by assuming the antecedent.                                  *)
  (*************************************************************************)
  <1>1. HAVE /\  LL1TypeInvariant
             /\  UniquenessInvariant
             /\  UNCHANGED LL1NVRAM
             /\  \A historyStateBinding \in HashType :
                     UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a new set of  *)
  (* variables in the appropriate types.  For the TAKE step to be          *)
  (* meaningful to the prover, first we have to tell the prover to expand  *)
  (* the definition of UniquenessInvariant, so it will see the universally *)
  (* quantified expression therein.                                        *)
  (*************************************************************************)
  <1> USE DEF UniquenessInvariant
  <1>2. TAKE stateHash1, stateHash2 \in HashType
  (*************************************************************************)
  (* To simplify the writing of the proof, we re-state the definitions     *)
  (* from the UniquenessInvariant.                                         *)
  (*************************************************************************)
  <1> historyStateBinding1 == Hash(LL1NVRAM.historySummary, stateHash1)
  <1> historyStateBinding2 == Hash(LL1NVRAM.historySummary, stateHash2)
  (*************************************************************************)
  (* We then assert the type safety of these definitions, with the help of *)
  (* the UniquenessInvariantDefsTypeSafeLemma.                             *)
  (*************************************************************************)
  <1>3. /\  historyStateBinding1 \in HashType
        /\  historyStateBinding2 \in HashType
    <2>1. LL1TypeInvariant
      BY <1>1
    <2>2. QED
      BY <1>2, <2>1, UniquenessInvariantDefsTypeSafeLemma
  (*************************************************************************)
  (* The UniquenessInvariant states an implication.  To prove this, it     *)
  (* suffices to assume the antecedent and prove the consequent.           *)
  (*************************************************************************)
  <1>4. SUFFICES
        ASSUME /\  LL1HistoryStateBindingAuthenticated(historyStateBinding1)'
               /\  LL1HistoryStateBindingAuthenticated(historyStateBinding2)'
        PROVE stateHash1 = stateHash2
    OBVIOUS
  (*************************************************************************)
  (* We hide the definitions, so they don't overwhelm the prover.  We'll   *)
  (* pull them in as necessary below.                                      *)
  (*************************************************************************)
  <1> HIDE DEF UniquenessInvariant
  <1> HIDE DEF historyStateBinding1, historyStateBinding2
  (*************************************************************************)
  (* First we'll show that history state binding 1 is authenticated in the *)
  (* unprimed state.                                                       *)
  (*************************************************************************)
  <1>5. LL1HistoryStateBindingAuthenticated(historyStateBinding1)
    (***********************************************************************)
    (* History state binding 1 is authenticated in the primed state, by    *)
    (* hypothesis.                                                         *)
    (***********************************************************************)
    <2>1. LL1HistoryStateBindingAuthenticated(historyStateBinding1)'
      BY <1>4
    (***********************************************************************)
    (* The authentication status of history state binding 1 has not        *)
    (* changed, because this status has not changed for any history state  *)
    (* binding in HashType.                                                *)
    (***********************************************************************)
    <2>2. UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding1)
      (*********************************************************************)
      (* We have to show that the history state binding has not changed,   *)
      (* which is not obvious because it is defined in terms of the        *)
      (* history summary in the NVRAM, so we have to derive this from the  *)
      (* fact that the NVRAM has not changed.                              *)
      (*********************************************************************)
      <3>1. UNCHANGED historyStateBinding1
        <4>1. UNCHANGED LL1NVRAM.historySummary
          <5>1. UNCHANGED LL1NVRAM
            BY <1>1
          <5>2. QED
            BY <5>1
        <4>2. QED
          BY <4>1 DEF historyStateBinding1
      (*********************************************************************)
      (* Then, we have to show that history state binding 1 has the        *)
      (* appropriate type.                                                 *)
      (*********************************************************************)
      <3>2. historyStateBinding1 \in HashType
        BY <1>3
      (*********************************************************************)
      (* The authentication status of all input state bindings has not     *)
      (* changed, as assumed by the lemma.                                 *)
      (*********************************************************************)
      <3>3. \A historyStateBinding \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
        BY <1>1
      (*********************************************************************)
      (* The conclusion follows directly.                                  *)
      (*********************************************************************)
      <3>4. QED
        BY <3>1, <3>2, <3>3
    (***********************************************************************)
    (* Since history state binding 1 is authenticated in the primed state, *)
    (* and since its authentication status has not changed, it is also     *)
    (* authenticated in the unprimed state.                                *)
    (***********************************************************************)
    <2>3. QED
      BY <2>1, <2>2
  (*************************************************************************)
  (* The same argument holds for history state binding 2.                  *)
  (*************************************************************************)
  <1>6. LL1HistoryStateBindingAuthenticated(historyStateBinding2)
    (***********************************************************************)
    (* History state binding 2 is authenticated in the primed state, by    *)
    (* hypothesis.                                                         *)
    (***********************************************************************)
    <2>1. LL1HistoryStateBindingAuthenticated(historyStateBinding2)'
      BY <1>4
    (***********************************************************************)
    (* The authentication status of history state binding 2 has not        *)
    (* changed, because this status has not changed for any history state  *)
    (* binding in HashType.                                                *)
    (***********************************************************************)
    <2>2. UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding2)
      (*********************************************************************)
      (* We have to show that the history state binding has not changed,   *)
      (* which is not obvious because it is defined in terms of the        *)
      (* history summary in the NVRAM, so we have to derive this from the  *)
      (* fact that the NVRAM has not changed.                              *)
      (*********************************************************************)
      <3>1. UNCHANGED historyStateBinding2
        <4>1. UNCHANGED LL1NVRAM.historySummary
          <5>1. UNCHANGED LL1NVRAM
            BY <1>1
          <5>2. QED
            BY <5>1
        <4>2. QED
          BY <4>1 DEF historyStateBinding2
      (*********************************************************************)
      (* Then, we have to show that history state binding 2 has the        *)
      (* appropriate type.                                                 *)
      (*********************************************************************)
      <3>2. historyStateBinding2 \in HashType
        BY <1>3
      (*********************************************************************)
      (* The authentication status of all input state bindings has not     *)
      (* changed, as assumed by the lemma.                                 *)
      (*********************************************************************)
      <3>3. \A historyStateBinding \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
        BY <1>1
      (*********************************************************************)
      (* The conclusion follows directly.                                  *)
      (*********************************************************************)
      <3>4. QED
        BY <3>1, <3>2, <3>3
    (***********************************************************************)
    (* Since history state binding 1 is authenticated in the primed state, *)
    (* and since its authentication status has not changed, it is also     *)
    (* authenticated in the unprimed state.                                *)
    (***********************************************************************)
    <2>3. QED
      BY <2>1, <2>2
  (*************************************************************************)
  (* Since the UniquenessInvariant holds in the unprimed state, it follows *)
  (* directly that the two state hashes are equal.                         *)
  (*************************************************************************)
  <1>7. QED
    <2>1. stateHash1 \in HashType
      BY <1>2
    <2>2. stateHash2 \in HashType
      BY <1>2
    <2>3. UniquenessInvariant
      BY <1>1
    <2>4. QED
      BY <1>5, <1>6, <2>1, <2>2, <2>3
      DEF UniquenessInvariant, historyStateBinding1, historyStateBinding2

----

(***************************************************************************)
(* The UnchangedAuthenticatedHistoryStateBindingsLemma states that if the  *)
(* NVRAM and the set of observed authenticators does not change, then      *)
(* there is no change to the set of history state bindings that have       *)
(* authenticators in the set LL1ObservedAuthenticators.                    *)
(***************************************************************************)

THEOREM UnchangedAuthenticatedHistoryStateBindingsLemma == \*`^\label{THEOREM-UnchangedAuthenticatedHistoryStateBindingsLemma}^'
    UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>> =>
        \A historyStateBinding \in HashType :
            UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
  (*************************************************************************)
  (* We assume the antecedents.                                            *)
  (*************************************************************************)
  <1>1. HAVE UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
  (*************************************************************************)
  (* To prove the universally quantified expression, we take a new history *)
  (* state binding in HashType.                                            *)
  (*************************************************************************)
  <1>2. TAKE historyStateBinding \in HashType
  (*************************************************************************)
  (* One fact that will be useful in several places is that the symmetric  *)
  (* key in the NVRAM has not changed.                                     *)
  (*************************************************************************)
  <1>3. UNCHANGED LL1NVRAM.symmetricKey
    <2>1. UNCHANGED LL1NVRAM
      BY <1>1
    <2>2. QED
      BY <2>1
  (*************************************************************************)
  (* We'll subdivide these into two cases.  In the first case, we'll       *)
  (* consider the history state bindings that are authenticated in the     *)
  (* unprimed state, and we'll show that they continue to be authenticated *)
  (* in the primed state.                                                  *)
  (*************************************************************************)
  <1>4. CASE LL1HistoryStateBindingAuthenticated(historyStateBinding) = TRUE
    <2>1. LL1HistoryStateBindingAuthenticated(historyStateBinding)' = TRUE
      (*********************************************************************)
      (* By hypothesis, the history state binding is authenticated in the  *)
      (* unprimed state.  Thus, we can pick an authenticator in the set of *)
      (* observed authenticators that is a valid MAC for this history      *)
      (* state binding.                                                    *)
      (*********************************************************************)
      <3>1. PICK authenticator \in LL1ObservedAuthenticators :
                ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, authenticator)
        BY <1>4 DEF LL1HistoryStateBindingAuthenticated
      (*********************************************************************)
      (* Because the symmetric key in the NVRAM has not changed, this      *)
      (* authenticator is also a valid MAC for this history state binding  *)
      (* in the primed state.                                              *)
      (*********************************************************************)
      <3>2. ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, authenticator)
        BY <1>3, <3>1
      (*********************************************************************)
      (* Because the set of observed authenticators has not changed, this  *)
      (* authenticator is also in the primed set of observed               *)
      (* authenticators.                                                   *)
      (*********************************************************************)
      <3>3. authenticator \in LL1ObservedAuthenticators'
        <4>1. authenticator \in LL1ObservedAuthenticators
          BY <3>1
        <4>2. UNCHANGED LL1ObservedAuthenticators
          BY <1>1
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The previous two conditions are sufficient to establish that the  *)
      (* history state binding is authenticated in the primed state.       *)
      (*********************************************************************)
      <3>4. QED
        BY <3>2, <3>3 DEF LL1HistoryStateBindingAuthenticated
    (***********************************************************************)
    (* Because the history state binding is authenticated in both the      *)
    (* unprimed and primed states, the LL1HistoryStateBindingAuthenticated *)
    (* is unchanged for this history state binding.                        *)
    (***********************************************************************)
    <2>2. QED
      BY <1>4, <2>1
  (*************************************************************************)
  (* In the second case, we'll consider the history state bindings that    *)
  (* are unauthenticated in the unprimed state, and we'll show that they   *)
  (* continue to be unauthenticated in the primed state.                   *)
  (*************************************************************************)
  <1>5. CASE LL1HistoryStateBindingAuthenticated(historyStateBinding) = FALSE
    <2>1. LL1HistoryStateBindingAuthenticated(historyStateBinding)' = FALSE
      (*********************************************************************)
      (* To prove that the history state binding is not authenticated in   *)
      (* the primed state, it suffices to show that none of the state      *)
      (* authenticators in the primed set of observed authenticators is a  *)
      (* valid MAC for the history state binding.                          *)
      (*********************************************************************)
      <3>1. SUFFICES \A authenticator \in LL1ObservedAuthenticators' :
                ~ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, authenticator)
        BY DEF LL1HistoryStateBindingAuthenticated
      (*********************************************************************)
      (* To prove the universally quantified expression, we take a new     *)
      (* authenticator in the primed set of observed authenticators.       *)
      (*********************************************************************)
      <3>2. TAKE authenticator \in LL1ObservedAuthenticators'
      (*********************************************************************)
      (* Because the set of observed authenticators has not changed, this  *)
      (* authenticator is also in the unprimed set of observed             *)
      (* authenticators.                                                   *)
      (*********************************************************************)
      <3>3. authenticator \in LL1ObservedAuthenticators
        <4>1. UNCHANGED LL1ObservedAuthenticators
          BY <1>1
        <4>2. QED
          BY <3>2, <4>1
      (*********************************************************************)
      (* Because the authenticator failed to authenticate the history      *)
      (* state binding in the unprimed state, and the symmetric key has    *)
      (* not changed, it immediately follows that the authenticator will   *)
      (* not authenticate the history state binding in the primed state.   *)
      (*********************************************************************)
      <3>4. QED
        BY <1>3, <1>5, <3>3 DEF LL1HistoryStateBindingAuthenticated
    (***********************************************************************)
    (* Because the history state binding is unauthenticated in both the    *)
    (* unprimed and primed states, the LL1HistoryStateBindingAuthenticated *)
    (* is unchanged for this history state binding.                        *)
    (***********************************************************************)
    <2>2. QED
      BY <1>5, <2>1
  (*************************************************************************)
  (* By proving that LL1HistoryStateBindingAuthenticated is a boolean      *)
  (* predicate, it is immediately clear that the two cases of true and     *)
  (* false are exhaustive for this predicate.                              *)
  (*************************************************************************)
  <1>6. LL1HistoryStateBindingAuthenticated(historyStateBinding) \in BOOLEAN
    BY DEF LL1HistoryStateBindingAuthenticated
  <1>7. QED
    BY <1>4, <1>5, <1>6

====
