---- MODULE MemoirLL1Specification ----

(***************************************************************************)
(* This module defines the low-level specification of `^\basicname^'.      *)
(*                                                                         *)
(* There are eight actions:                                                *)
(*     LL1MakeInputAvailable                                               *)
(*     LL1PerformOperation                                                 *)
(*     LL1RepeatOperation                                                  *)
(*     LL1Restart                                                          *)
(*     LL1ReadDisk                                                         *)
(*     LL1WriteDisk                                                        *)
(*     LL1CorruptRAM                                                       *)
(*     LL1RestrictedCorruption                                             *)
(***************************************************************************)

EXTENDS MemoirLLPrimitives

(***************************************************************************)
(* The disk and RAM are untrusted.  They each can store the service's      *)
(* public state, encrypted private state, a history summary (a chained     *)
(* hash of all inputs that the service has processed), and an              *)
(* authenticator (a MAC that binds the history summary to the public and   *)
(* private state).                                                         *)
(***************************************************************************)

LL1UntrustedStorageType == [ \*`^\label{DEF-LL1UntrustedStorageType}^'
    publicState : PublicStateType,
    privateStateEnc : PrivateStateEncType,
    historySummary : HashType,
    authenticator : MACType ]

(***************************************************************************)
(* The NVRAM is trusted, since the TPM guarantees that it can only be read *)
(* or written by the code that implements `^\name^'.  The NVRAM stores the *)
(* current history summary and the symmetric key that is used (1) to       *)
(* encrypt the private state and (2) to MAC the history summary and        *)
(* service state into an authenticator.                                    *)
(***************************************************************************)

LL1TrustedStorageType == [ \*`^\label{DEF-LL1TrustedStorageType}^'
    historySummary : HashType,
    symmetricKey : SymmetricKeyType ]

(***************************************************************************)
(* LL1AvailableInputs and LL1ObservedOutputs are abstract variables that   *)
(* do not directly represent part of the implementation.  They correspond  *)
(* to the HLAvailableInputs and HLObservedOutputs variables in the         *)
(* high-level spec.                                                        *)
(***************************************************************************)

VARIABLE LL1AvailableInputs \*`^\label{VARIABLE-LL1AvailableInputs}^'
VARIABLE LL1ObservedOutputs \*`^\label{VARIABLE-LL1ObservedOutputs}^'

(***************************************************************************)
(* The LL1ObservedAuthenticators variable is also abstract.  It records    *)
(* the set of authenticators that the user has seen from `^\name^'.  A     *)
(* malicious user can attempt to use these state authenticators in a      *)
(* replay attack against `^\name^'.                                        *)
(***************************************************************************)

VARIABLE LL1ObservedAuthenticators \*`^\label{VARIABLE-LL1ObservedAuthenticators}^'

(***************************************************************************)
(* The LL1Disk, LL1RAM, and LL1NVRAM variables represent concrete state    *)
(* maintained by the `^\basicname^' implementation.                        *)
(***************************************************************************)

VARIABLE LL1Disk \*`^\label{VARIABLE-LL1Disk}^'
VARIABLE LL1RAM \*`^\label{VARIABLE-LL1RAM}^'
VARIABLE LL1NVRAM \*`^\label{VARIABLE-LL1NVRAM}^'

LL1TypeInvariant == \*`^\label{DEF-LL1TypeInvariant}^'
    /\  LL1AvailableInputs \subseteq InputType
    /\  LL1ObservedOutputs \subseteq OutputType
    /\  LL1ObservedAuthenticators \subseteq MACType
    /\  LL1Disk \in LL1UntrustedStorageType
    /\  LL1RAM \in LL1UntrustedStorageType
    /\  LL1NVRAM \in LL1TrustedStorageType

LL1Vars == << \*`^\label{DEF-LL1Vars}^'
    LL1AvailableInputs,
    LL1ObservedOutputs,
    LL1ObservedAuthenticators,
    LL1Disk,
    LL1RAM,
    LL1NVRAM >>

----

(***************************************************************************)
(* The LL1MakeInputAvailable action is a direct analog of the              *)
(* HLMakeInputAvailable action in the high-level spec.                     *)
(***************************************************************************)

LL1MakeInputAvailable == \*`^\label{DEF-LL1MakeInputAvailable}^'
    \E input \in InputType :
        /\  input \notin LL1AvailableInputs
        /\  LL1AvailableInputs' = LL1AvailableInputs \union {input}
        /\  UNCHANGED LL1Disk
        /\  UNCHANGED LL1RAM
        /\  UNCHANGED LL1NVRAM
        /\  UNCHANGED LL1ObservedOutputs
        /\  UNCHANGED LL1ObservedAuthenticators

----

(***************************************************************************)
(* The LL1PerformOperation action is invoked by the user to perform a      *)
(* service operation.  It is intended to provide the semantics of the      *)
(* HLAdvanceService action in the high-level spec.                         *)
(***************************************************************************)

LL1PerformOperation == \*`^\label{DEF-LL1PerformOperation}^'
    \E input \in LL1AvailableInputs :
        LET
            stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
            historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
            privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
            sResult == Service(LL1RAM.publicState, privateState, input)
            newPrivateStateEnc ==
                SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
            newHistorySummary == Hash(LL1NVRAM.historySummary, input)
            newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
            newHistoryStateBinding == Hash(newHistorySummary, newStateHash)
            newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
        IN
                (***********************************************************)
                (* There are two enablement conditions: First, the         *)
                (* authenticator supplied by the user                      *)
                (* (LL1RAM.authenticator) must validly bind the            *)
                (* user-supplied public and encrypted private state to the *)
                (* history summary supplied by the user.  Second, the      *)
                (* history summary supplied by the user must match the     *)
                (* history summary in the NVRAM.                           *)
                (***********************************************************)
            /\  ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
            /\  LL1NVRAM.historySummary = LL1RAM.historySummary
                (***********************************************************)
                (* At the conclusion of the action, the RAM contains the   *)
                (* new public and encrypted private state, the new history *)
                (* summary, and an authenticator that binds these          *)
                (* together.                                               *)
                (***********************************************************)
            /\  LL1RAM' = [
                    publicState |-> sResult.newPublicState,
                    privateStateEnc |-> newPrivateStateEnc,
                    historySummary |-> newHistorySummary,
                    authenticator |-> newAuthenticator ]
                (***********************************************************)
                (* The NVRAM is updated with the new history summary.      *)
                (***********************************************************)
            /\  LL1NVRAM' = [
                    historySummary |-> newHistorySummary,
                    symmetricKey |-> LL1NVRAM.symmetricKey ]
                (***********************************************************)
                (* The output of the service is added to the set of        *)
                (* outputs that the user has observed.                     *)
                (***********************************************************)
            /\  LL1ObservedOutputs' = LL1ObservedOutputs \union {sResult.output}
                (***********************************************************)
                (* The disk is unchanged.                                  *)
                (***********************************************************)
            /\  UNCHANGED LL1Disk
                (***********************************************************)
                (* The set of available inputs is unchanged                *)
                (***********************************************************)
            /\  UNCHANGED LL1AvailableInputs
                (***********************************************************)
                (* The new authenticator is added to the set of            *)
                (* authenticators that the user has observed.              *)
                (***********************************************************)
            /\  LL1ObservedAuthenticators' =
                    LL1ObservedAuthenticators \union {newAuthenticator}

----

(***************************************************************************)
(* The LL1RepeatOperation action is invoked by the user when the computer  *)
(* crashed after a LL1PerformOperation action was performed but before the *)
(* user had a chance to perform a LL1WriteDisk action to persistently      *)
(* record the new state and its authenticator.  This action enables the    *)
(* user to reproduce the result of the most-recent LL1PerformOperation     *)
(* action.                                                                 *)
(***************************************************************************)

LL1RepeatOperation == \*`^\label{DEF-LL1RepeatOperation}^'
    \E input \in LL1AvailableInputs :
        LET
            stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
            historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
            privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
            sResult == Service(LL1RAM.publicState, privateState, input)
            newPrivateStateEnc ==
                SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
            newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
            newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
            newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
        IN
                (***********************************************************)
                (* There are two enablement conditions: First, the         *)
                (* authenticator supplied by the user                      *)
                (* (LL1RAM.authenticator) must validly bind the            *)
                (* user-supplied public and encrypted private state to the *)
                (* history summary supplied by the user.                   *)
                (*                                                         *)
                (* Second, the history summary supplied by the user,       *)
                (* hashed with the input supplied by the user, must match  *)
                (* the history summary in the NVRAM.  This condition       *)
                (* ensures that this action will invoke the service with   *)
                (* the same input used with the most recent                *)
                (* LL1PerformOperation action.                             *)
                (***********************************************************)
            /\  ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
            /\  LL1NVRAM.historySummary = Hash(LL1RAM.historySummary, input)
                (***********************************************************)
                (* At the conclusion of the action, the RAM contains the   *)
                (* new public and encrypted private state, the new history *)
                (* summary, and an authenticator that binds these          *)
                (* together.  These should match the values produced by    *)
                (* the most recent LL1PerformOperation action.             *)
                (***********************************************************)
            /\  LL1RAM' = [
                    publicState |-> sResult.newPublicState,
                    privateStateEnc |-> newPrivateStateEnc,
                    historySummary |-> LL1NVRAM.historySummary,
                    authenticator |-> newAuthenticator ]
                (***********************************************************)
                (* The output of the service is added to the set of        *)
                (* outputs that the user has observed.  If `^\name^' is    *)
                (* working correctly, the user already saw this output     *)
                (* when the previous LL1PerformOperation action was        *)
                (* executed.                                               *)
                (***********************************************************)
            /\  LL1ObservedOutputs' = LL1ObservedOutputs \union {sResult.output}
                (***********************************************************)
                (* The NVRAM is unchanged, because this action is not      *)
                (* supposed to change the state of the service.            *)
                (***********************************************************)
            /\  UNCHANGED LL1NVRAM
                (***********************************************************)
                (* The disk is unchanged.                                  *)
                (***********************************************************)
            /\  UNCHANGED LL1Disk
                (***********************************************************)
                (* The set of available inputs is unchanged                *)
                (***********************************************************)
            /\  UNCHANGED LL1AvailableInputs
                (***********************************************************)
                (* The new authenticator is added to the set of            *)
                (* authenticators that the user has observed.  If          *)
                (* `^\name^' is working correctly, the user already saw    *)
                (* this authenticator when the previous                    *)
                (* LL1PerformOperation action was executed.                *)
                (***********************************************************)
            /\  LL1ObservedAuthenticators' =
                    LL1ObservedAuthenticators \union {newAuthenticator}

----

(***************************************************************************)
(* The LL1Restart action occurs when the computer restarts.                *)
(***************************************************************************)

LL1Restart == \*`^\label{DEF-LL1Restart}^'
    \E  untrustedStorage \in LL1UntrustedStorageType,
        randomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
        hash \in HashType :
            (***************************************************************)
            (* The state of the RAM is trashed by a restart, so we set it  *)
            (* to some *almost* arbitrary value in                         *)
            (* LL1UntrustedStorageType.  The only condition we impose is   *)
            (* that the authenticator is not coincidentally equal to an    *)
            (* authenticator that could be computed with the symmetric key *)
            (* known only to `^\name^'.                                    *)
            (***************************************************************)
        /\  untrustedStorage.authenticator = GenerateMAC(randomSymmetricKey, hash)
        /\  LL1RAM' = untrustedStorage
        /\  UNCHANGED LL1Disk
        /\  UNCHANGED LL1NVRAM
        /\  UNCHANGED LL1AvailableInputs
        /\  UNCHANGED LL1ObservedOutputs
        /\  UNCHANGED LL1ObservedAuthenticators

----

(***************************************************************************)
(* The LL1ReadDisk action copies the state of the disk into the RAM.       *)
(***************************************************************************)

LL1ReadDisk == \*`^\label{DEF-LL1ReadDisk}^'
    /\  LL1RAM' = LL1Disk
    /\  UNCHANGED LL1Disk
    /\  UNCHANGED LL1NVRAM
    /\  UNCHANGED LL1AvailableInputs
    /\  UNCHANGED LL1ObservedOutputs
    /\  UNCHANGED LL1ObservedAuthenticators

----

(***************************************************************************)
(* The LL1WriteDisk action copies the state of the RAM onto the disk.      *)
(***************************************************************************)

LL1WriteDisk == \*`^\label{DEF-LL1WriteDisk}^'
    /\  LL1Disk' = LL1RAM
    /\  UNCHANGED LL1RAM
    /\  UNCHANGED LL1NVRAM
    /\  UNCHANGED LL1AvailableInputs
    /\  UNCHANGED LL1ObservedOutputs
    /\  UNCHANGED LL1ObservedAuthenticators

----

(***************************************************************************)
(* The LL1CorruptRAM action models the ability of a malicious user to      *)
(* attack `^\name^' by supplying *almost* arbitrary data to `^\name^'.     *)
(* The data is not completely arbitrary, because the user is assumed to be *)
(* unable to forge authenticators using the symmetric key stored in the    *)
(* NVRAM of `^\name^'.                                                     *)
(***************************************************************************)

LL1CorruptRAM == \*`^\label{DEF-LL1CorruptRAM}^'
    \E  untrustedStorage \in LL1UntrustedStorageType,
        fakeSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
        hash \in HashType :
                (***********************************************************)
                (* The user can launch a replay attack by re-using any     *)
                (* authenticator previously observed.                      *)
                (***********************************************************)
        /\  \/  untrustedStorage.authenticator \in LL1ObservedAuthenticators
                (***********************************************************)
                (* Or the user can create a fake authenticator using some  *)
                (* other symmetric key.                                    *)
                (***********************************************************)
            \/  untrustedStorage.authenticator = GenerateMAC(fakeSymmetricKey, hash)
        /\  LL1RAM' = untrustedStorage
        /\  UNCHANGED LL1Disk
        /\  UNCHANGED LL1NVRAM
        /\  UNCHANGED LL1AvailableInputs
        /\  UNCHANGED LL1ObservedOutputs
        /\  UNCHANGED LL1ObservedAuthenticators

----

(***************************************************************************)
(* The LL1RestrictedCorruption does not model any realistic action in a    *)
(* direct implementation of this low-level spec.  The TPM prevents any     *)
(* code other than `^\name^' from writing to the NVRAM, so an attacker     *)
(* cannot actually perform this action in `^\basicname^'.                  *)
(*                                                                         *)
(* However, `^\fastname^' allows an attacker to perform an action          *)
(* (LL2CorruptSPCR) that corrupts the stored history summary in            *)
(* `^\fastname^', and in the correctness proof of `^\fastname^', we will   *)
(* show that the CorruptSPCR action in `^\fastname^' (under some           *)
(* circumstances) refines to the LL1RestrictedCorruption action in         *)
(* `^\basicname^'.                                                         *)
(*                                                                         *)
(* Similarly, when a LL2Restart action occurs in `^\fastname^', there are  *)
(* circumstances that cause this to refine to the LL1RestrictedCorruption  *)
(* action in `^\basicname^'.                                               *)
(*                                                                         *)
(* For this reason, we need the LL1RestrictedCorruption action to be       *)
(* strong enough to enable refinement from the LL2CorruptSPCR and          *)
(* LL2Restart actions in `^\fastname^', but weak enough to enable          *)
(* refinement to an action in the high-level spec, specifically the HLDie  *)
(* action.                                                                 *)
(*                                                                         *)
(* We therefore impose a somewhat bizarre-looking pair of constraints on   *)
(* the garbage value to which an attacker can set the history summary in   *)
(* the NVRAM.  The first constraint (labeled `current') is needed to       *)
(* ensure that the CardinalityInvariant and UniquenessInvariant continue   *)
(* to hold when a LL1RestrictedCorruption action occurs, and the second    *)
(* constraint (labeled `previous') is needed to ensure that the            *)
(* InclusionInvariant continues to hold when a LL1RestrictedCorruption     *)
(* action occurs.                                                          *)
(***************************************************************************)

LL1RestrictedCorruption == \*`^\label{DEF-LL1RestrictedCorruption}^'
    /\  nvram::
        \E garbageHistorySummary \in HashType :
            /\  current(garbageHistorySummary)::
                (***********************************************************)
                (* There is no authenticator that validates a history      *)
                (* state binding that binds the the garbage history        *)
                (* summary to any state hash.                              *)
                (***********************************************************)
                \A  stateHash \in HashType, authenticator \in LL1ObservedAuthenticators :
                    LET
                        historyStateBinding == Hash(garbageHistorySummary, stateHash)
                    IN
                        ~ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, authenticator)
            /\  previous(garbageHistorySummary)::
                (***********************************************************)
                (* There is no authenticator that validates a history      *)
                (* state binding that binds any predecessor of the the     *)
                (* garbage history summary to any state hash.              *)
                (***********************************************************)
                \A  stateHash \in HashType,
                    authenticator \in LL1ObservedAuthenticators,
                    someHistorySummary \in HashType,
                    someInput \in InputType :
                    LET
                        historyStateBinding == Hash(someHistorySummary, stateHash)
                    IN
                        garbageHistorySummary = Hash(someHistorySummary, someInput) =>
                            ~ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, authenticator)
                (***********************************************************)
                (* The history summary in the NVRAM becomes equal to the    *)
                (* garbage history summary.                                 *)
                (***********************************************************)
            /\  LL1NVRAM' = [
                    historySummary |-> garbageHistorySummary,
                    symmetricKey |-> LL1NVRAM.symmetricKey ]
    /\  ram::
        \/  unchanged::
            (***************************************************************)
            (* A LL2CorruptSPCR action in the `^\fastname^' spec leaves    *)
            (* the state of the RAM unchanged.                             *)
            (***************************************************************)
            UNCHANGED LL1RAM
        \/  trashed::
            (***************************************************************)
            (* A LL2Restart action in the `^\fastname^' spec trashes the   *)
            (* RAM, so we set it to some *almost* arbitrary value in       *)
            (* LL1UntrustedStorageType.  The only condition we impose is   *)
            (* that the authenticator is not coincidentally equal to an    *)
            (* authenticator that could be computed with the symmetric key *)
            (* known only to `^\name^'.                                    *)
            (***************************************************************)
            \E  untrustedStorage \in LL1UntrustedStorageType,
                randomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                hash \in HashType :
                /\  untrustedStorage.authenticator = GenerateMAC(randomSymmetricKey, hash)
                /\  LL1RAM' = untrustedStorage
    /\  UNCHANGED LL1Disk
    /\  UNCHANGED LL1AvailableInputs
    /\  UNCHANGED LL1ObservedOutputs
    /\  UNCHANGED LL1ObservedAuthenticators

----

(***************************************************************************)
(* In the initial state of the `^\basicname^' implementation, some         *)
(* symmetric key is generated and stored in the NVRAM.  The initial        *)
(* history summary is the base hash value, indicating that no inputs have  *)
(* been supplied yet.  An initial authenticator binds the initial history  *)
(* summary to the initial public and encrypted private state.              *)
(***************************************************************************)

LL1Init == \*`^\label{DEF-LL1Init}^'
    \E symmetricKey \in SymmetricKeyType :
        LET
            initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
            initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
            initialHistoryStateBinding == Hash(BaseHashValue, initialStateHash)
            initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
            initialUntrustedStorage == [
                publicState |-> InitialPublicState,
                privateStateEnc |-> initialPrivateStateEnc,
                historySummary |-> BaseHashValue,
                authenticator |-> initialAuthenticator ]
            initialTrustedStorage == [
                historySummary |-> BaseHashValue,
                symmetricKey |-> symmetricKey ]
        IN
            /\  LL1Disk = initialUntrustedStorage
            /\  LL1RAM = initialUntrustedStorage
            /\  LL1NVRAM = initialTrustedStorage
            /\  LL1AvailableInputs = InitialAvailableInputs
            /\  LL1ObservedOutputs = {}
            /\  LL1ObservedAuthenticators = {initialAuthenticator}

LL1Next == \*`^\label{DEF-LL1Next}^'
    \/  LL1MakeInputAvailable
    \/  LL1PerformOperation
    \/  LL1RepeatOperation
    \/  LL1Restart
    \/  LL1ReadDisk
    \/  LL1WriteDisk
    \/  LL1CorruptRAM
    \/  LL1RestrictedCorruption

LL1Spec == LL1Init /\ [][LL1Next]_LL1Vars \*`^\label{DEF-LL1Spec}^'

====
