---- MODULE MemoirLL1SupplementalInvariants ----

(***************************************************************************)
(* This module defines two supplemental invariants.  These are not needed  *)
(* directly by the `^\basicname^' implementation proof, but they are       *)
(* needed by the proofs that the correctness invariants hold.              *)
(***************************************************************************)

EXTENDS MemoirLL1CorrectnessInvariants, Naturals

----

(***************************************************************************)
(* The ExtendedUnforgeabilityInvariant states that the unforgeability      *)
(* property of the authenticator in the RAM also applies to the            *)
(* authenticator on the disk.  This is needed to show that the             *)
(* UnforgeabilityInvariant holds through a LL1ReadDisk action.             *)
(***************************************************************************)

ExtendedUnforgeabilityInvariant == \*`^\label{DEF-ExtendedUnforgeabilityInvariant}^'
    \A historyStateBinding \in HashType :
        /\  ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator) =>
                LL1RAM.authenticator \in LL1ObservedAuthenticators
        /\  ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1Disk.authenticator) =>
                LL1Disk.authenticator \in LL1ObservedAuthenticators

----

(***************************************************************************)
(* The CardinalityInvariant is a little funky.  We define a new operatior  *)
(* called HashCardinality that indicates the count of hashes needed to     *)
(* produce the supplied hash value.  For example, if you create a hash     *)
(* chain, starting with the base hash value, and chain in N inputs, the    *)
(* cardinality of the resulting hash will be N.  Although our low-level    *)
(* spec uses the hash operator only for linear hash chains, the hash       *)
(* cardinality is also defined for arbitrary trees of hashes.              *)
(*                                                                         *)
(* The CardinalityInvariant states that the hash cardinality of the        *)
(* history summary in any observed authenticator is less than or equal to  *)
(* the hash cardinality of the history summary in the NVRAM.               *)
(*                                                                         *)
(* This property is needed to prove that the uniqueness property is an     *)
(* inductive invariant.                                                    *)
(***************************************************************************)

CONSTANT HashCardinality(_) \*`^\label{CONSTANT-HashCardinality}^'

ASSUME HashCardinalityTypeSafe == \*`^\label{ASSUME-HashCardinalityTypeSafe}^'
    \A hash \in HashDomain : HashCardinality(hash) \in Nat

ASSUME BaseHashCardinalityZero == HashCardinality(BaseHashValue) = 0 \*`^\label{ASSUME-BaseHashCardinalityZero}^'

ASSUME InputCardinalityZero == \*`^\label{ASSUME-InputCardinalityZero}^'
    \A input \in InputType : HashCardinality(input) = 0

ASSUME HashCardinalityAccumulative == \*`^\label{ASSUME-HashCardinalityAccumulative}^'
    \A hash1, hash2 \in HashDomain :
        HashCardinality(Hash(hash1, hash2)) =
            HashCardinality(hash1) + HashCardinality(hash2) + 1

CardinalityInvariant == \*`^\label{DEF-CardinalityInvariant}^'
    \A historySummary \in HashType, stateHash \in HashType :
        LET
            historyStateBinding == Hash(historySummary, stateHash)
        IN
            ( /\  LL1NVRAMHistorySummaryUncorrupted
              /\  LL1HistoryStateBindingAuthenticated(historyStateBinding) )
          =>
              HashCardinality(historySummary) <= HashCardinality(LL1NVRAM.historySummary)

====
