---- MODULE MemoirLLPrimitives ----

(***************************************************************************)
(* This module defines primitives that are used by the low-level specs.    *)
(* The primitives include a hash function, MAC functions, and symmetric    *)
(* crypto functions, along with their associated types.  The module also   *)
(* asserts assumptions about the properties of these functions.            *)
(***************************************************************************)

EXTENDS MemoirHLTypeSafety

(***************************************************************************)
(* The low-level specs make use of three primitives: a secure hash, a MAC  *)
(* (message authentication code), and symmetric cryptography.              *)
(***************************************************************************)

CONSTANT HashType \*`^\label{CONSTANT-HashType}^'
CONSTANT MACType \*`^\label{CONSTANT-MACType}^'
CONSTANT SymmetricKeyType \*`^\label{CONSTANT-SymmetricKeyType}^'
CONSTANT PrivateStateEncType \*`^\label{CONSTANT-PrivateStateEncType}^'

CONSTANT Hash(_,_) \*`^\label{CONSTANT-Hash}^'
CONSTANT GenerateMAC(_,_) \*`^\label{CONSTANT-GenerateMAC}^'
CONSTANT ValidateMAC(_,_,_) \*`^\label{CONSTANT-ValidateMAC}^'
CONSTANT SymmetricEncrypt(_,_) \*`^\label{CONSTANT-SymmetricEncrypt}^'
CONSTANT SymmetricDecrypt(_,_) \*`^\label{CONSTANT-SymmetricDecrypt}^'

(***************************************************************************)
(* The hash function has a somewhat strange signature.  It accepts two     *)
(* arguments, rather than one.  The reason for this is so that we can      *)
(* construct hash chains.  Alternatively, we could have written the spec   *)
(* with a conventional single-argument hash function and a two-argument    *)
(* concatenation function, but this would have added complexity for no     *)
(* real benefit.                                                           *)
(***************************************************************************)

(***************************************************************************)
(* We assume a base hash value, which in a real implementation, might just *)
(* be the value zero.                                                      *)
(***************************************************************************)

CONSTANT BaseHashValue \*`^\label{CONSTANT-BaseHashValue}^'

(***************************************************************************)
(* The domain of the hash function is hashes, inputs, public states, and   *)
(* encrypted private states.  These are the only types we need to hash.    *)
(***************************************************************************)

HashDomain == UNION { \*`^\label{DEF-HashDomain}^'
    HashType,
    InputType,
    PublicStateType,
    PrivateStateEncType }

(***************************************************************************)
(* The base hash value is a valid hash value, and it cannot be produced by *)
(* hashing any other value.                                                *)
(***************************************************************************)

ASSUME BaseHashValueTypeSafe == BaseHashValue \in HashType \*`^\label{ASSUME-BaseHashValueTypeSafe}^'

ASSUME BaseHashValueUnique == \*`^\label{ASSUME-BaseHashValueUnique}^'
    \A hashInput1, hashInput2 \in HashDomain :
        Hash(hashInput1, hashInput2) # BaseHashValue

(***************************************************************************)
(* The hash function is assumed to be type-safe and collision-resistant.   *)
(* In this spec, we define collision resistance in a very strong sense,    *)
(* namely that there are no different inputs that will hash to the same    *)
(* output.  Although a real implementation of a hash function cannot       *)
(* satisfy this, cryptographically secure hash functions are expected to   *)
(* practically satisfy such a condition.                                   *)
(***************************************************************************)

ASSUME HashTypeSafe == \*`^\label{ASSUME-HashTypeSafe}^'
    \A hashInput1, hashInput2 \in HashDomain : Hash(hashInput1, hashInput2) \in HashType

ASSUME HashCollisionResistant == \*`^\label{ASSUME-HashCollisionResistant}^'
    \A hashInput1a, hashInput2a, hashInput1b, hashInput2b \in HashDomain :
        Hash(hashInput1a, hashInput2a) = Hash(hashInput1b, hashInput2b) =>
            /\  hashInput1a = hashInput1b
            /\  hashInput2a = hashInput2b

(***************************************************************************)
(* The MAC functions are assumed to be type-safe, complete, consistent,    *)
(* unforgeable, and collision-resistant.                                   *)
(***************************************************************************)

ASSUME GenerateMACTypeSafe == \*`^\label{ASSUME-GenerateMACTypeSafe}^'
    \A key \in SymmetricKeyType, hash \in HashType :
        GenerateMAC(key, hash) \in MACType

ASSUME ValidateMACTypeSafe == \*`^\label{ASSUME-ValidateMACTypeSafe}^'
    \A key \in SymmetricKeyType, hash \in HashType, mac \in MACType :
        ValidateMAC(key, hash, mac) \in BOOLEAN

ASSUME MACComplete == \*`^\label{ASSUME-MACComplete}^'
    \A key \in SymmetricKeyType, hash \in HashType :
        ValidateMAC(key, hash, GenerateMAC(key, hash)) = TRUE

ASSUME MACConsistent == \*`^\label{ASSUME-MACConsistent}^'
    \A key \in SymmetricKeyType, hash \in HashType, mac \in MACType :
        ValidateMAC(key, hash, mac) => mac = GenerateMAC(key, hash)

ASSUME MACUnforgeable == \*`^\label{ASSUME-MACUnforgeable}^'
    \A key1, key2 \in SymmetricKeyType, hash1, hash2 \in HashType :
        ValidateMAC(key1, hash1, GenerateMAC(key2, hash2)) => key1 = key2

ASSUME MACCollisionResistant == \*`^\label{ASSUME-MACCollisionResistant}^'
    \A key1, key2 \in SymmetricKeyType, hash1, hash2 \in HashType :
        ValidateMAC(key1, hash1, GenerateMAC(key2, hash2)) => hash1 = hash2

(***************************************************************************)
(* The symmetric-crypto functions are assumed to be type-safe.  They are   *)
(* also assumed to be correct, meaning that decryption is the inverse of   *)
(* encryption, given the same crypto key.                                  *)
(***************************************************************************)

ASSUME SymmetricEncryptionTypeSafe == \*`^\label{ASSUME-SymmetricEncryptionTypeSafe}^'
    \A key \in SymmetricKeyType, privateState \in PrivateStateType :
        SymmetricEncrypt(key, privateState) \in PrivateStateEncType

ASSUME SymmetricDecryptionTypeSafe == \*`^\label{ASSUME-SymmetricDecryptionTypeSafe}^'
    \A key \in SymmetricKeyType, privateStateEnc \in PrivateStateEncType :
        SymmetricDecrypt(key, privateStateEnc) \in PrivateStateType

ASSUME SymmetricCryptoCorrect == \*`^\label{ASSUME-SymmetricCryptoCorrect}^'
    \A key \in SymmetricKeyType, privateState \in PrivateStateType :
        SymmetricDecrypt(key, SymmetricEncrypt(key, privateState)) = privateState

====
